from __future__ import annotations

import dataclasses
import json
import sys
from dataclasses import is_dataclass, fields
from enum import Enum

from types import GenericAlias, GeneratorType

try:
    from types import UnionType, NoneType
except ImportError:
    # python 3.9
    from typing import Union

    UnionType = type(Union)
    NoneType = None

from typing import Any, ForwardRef
from typing import _UnionGenericAlias, _eval_type

from PySide2 import QtCore, QtGui


def cast(typ: Any, value: Any, globalns: dict | None = None) -> Any:
    """Casts a value to a type or a type hint.

    Raises TypeError when failed.
    """

    if globalns is None:
        globalns = {}

    if typ in (None, NoneType):
        return None

    elif typ is Any:
        return value

    elif isinstance(typ, type) and isinstance(value, typ):
        return value

    elif isinstance(typ, str):
        # check if string is a ForwardRef, such as 'InventoryItem'
        typ = _eval_type(ForwardRef(typ), globalns, globalns)
        return cast(typ, value)

    elif isinstance(typ, type) and issubclass(typ, (list, tuple, set)):
        return typ(value)

    elif isinstance(typ, type) and issubclass(typ, Enum):
        # type hints such as enum.Enum
        if isinstance(value, typ):
            # value is Enum
            return value
        try:
            enum = typ[value]
            return enum
        except KeyError:
            return len(typ) and list(typ)[0] or None

    elif isinstance(typ, GenericAlias):
        # type hints such as list[int], dict[str, int] etc.
        origin = typ.__origin__
        args = typ.__args__
        if issubclass(origin, list):
            try:
                arg = args[0]
                return origin(cast(arg, v, globalns) for v in value)
            except IndexError:
                return origin(value)
            except TypeError:
                return value

        elif issubclass(origin, tuple):
            if len(args) == 2 and args[1] is Ellipsis:
                return origin(cast(args[0], v, globalns) for v in value)
            try:
                return origin(cast(args[i], v, globalns) for i, v in enumerate(value))
            except IndexError:
                return origin(value)

        elif issubclass(origin, dict):
            try:
                key_arg = args[0]
                value_arg = args[1]
                return origin(
                    (cast(key_arg, k, globalns), cast(value_arg, v, globalns))
                    for k, v in value.items()
                )
            except IndexError:
                return dict(**value)

    elif isinstance(typ, (UnionType, _UnionGenericAlias)):
        # type hints such as int | str, Optional[int]

        if type(value) in typ.__args__:
            # don't convert if current type comes later in list
            return value
        # iterate through all types until one works
        for arg in typ.__args__:
            try:
                return cast(arg, value, globalns)
            except (ValueError, TypeError):
                continue

    elif isinstance(typ, ForwardRef):
        typ = _eval_type(typ, globalns, globalns)
        return cast(typ, value)

    elif is_dataclass(typ):
        # types hints that are dataclasses

        # update globals (see typing module)
        base_globals = getattr(sys.modules.get(typ.__module__, None), '__dict__', {})
        globalns.update(base_globals)

        kwargs = {}
        if isinstance(value, dict):
            # treat value as kwargs for dataclass
            for field_ in fields(typ):
                if field_.name not in value:
                    continue
                kw_value = cast(field_.type, value[field_.name], globalns)
                kwargs[field_.name] = kw_value
        elif isinstance(value, (list, tuple)):
            # treat value as *args for a dataclass
            for field_, v in zip(fields(typ), value):
                kw_value = cast(field_.type, v, globalns)
                kwargs[field_.name] = kw_value
        else:
            return typ(value)
        return typ(**kwargs)

    else:
        # all other type hints
        if isinstance(value, (tuple, list)):
            return typ(*value)
        elif isinstance(value, Enum):
            return typ(value.value)
        else:
            return typ(value)

    raise TypeError(f'cannot cast to type ({typ})')


def basic(obj: Any) -> Any:
    """Returns a basic type that can be serialized by json."""

    if dataclasses.is_dataclass(obj):
        obj = dataclasses.asdict(obj)

    if obj is None:
        return None
    if isinstance(obj, (str, int, float, bool)):
        # basic types as defined in json library
        return obj
    elif isinstance(obj, (list, GeneratorType)):
        return type(obj)(basic(v) for v in obj)
    elif isinstance(obj, dict):
        return type(obj)((k, basic(v)) for k, v in obj.items())
    elif isinstance(obj, tuple):
        return type(obj)(basic(v) for v in obj)
    elif isinstance(obj, set):
        return [basic(v) for v in obj]
    elif isinstance(obj, Enum):
        return obj.name
    elif isinstance(obj, (QtCore.QPoint, QtCore.QPointF)):
        return [obj.x(), obj.y()]
    elif isinstance(obj, (QtCore.QSize, QtCore.QSizeF)):
        return [obj.width(), obj.height()]
    elif isinstance(obj, (QtCore.QRect, QtCore.QRectF)):
        return [obj.x(), obj.y(), obj.width(), obj.height()]
    elif isinstance(obj, QtGui.QColor):
        return obj.getRgb()

    try:
        # convert enum, flags to int
        if int(obj) == obj:
            return int(obj)
    except TypeError:
        pass

    raise TypeError(f'Cannot convert {obj.__class__} to basic type.')


def deep_field(obj) -> dataclasses.Field:
    return dataclasses.field(default_factory=lambda: obj.__class__(obj))


def hashable_dataclass(cls):
    def __hash__(self) -> int:
        if self._hash is None:
            data = basic(self)
            _hash = hash(json.dumps(data, sort_keys=True))
            super(cls, self).__setattr__('_hash', _hash)
        return self._hash

    def __setattr__(self, name: str, value: Any) -> None:
        super(cls, self).__setattr__(name, value)
        super(cls, self).__setattr__('_hash', None)

    cls._hash = None
    cls.__setattr__ = __setattr__
    cls.__hash__ = __hash__
    return dataclasses.dataclass(cls)


class HashableDict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))



from __future__ import annotations

import dataclasses
import logging
import os
import shutil

from PySide2 import QtWidgets, QtCore, QtGui

from qt_extensions.helper import unique_path
from qt_extensions.icons import MaterialIcon
from qt_extensions.elementbrowser import ElementBrowser, Field, ElementDelegate

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class FileElement:
    name: str
    path: str


class FileNameDelegate(ElementDelegate):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        file_regex = QtCore.QRegularExpression(r'[\w\d\.-]+')
        self.validator = QtGui.QRegularExpressionValidator(file_regex, self)

    def createEditor(
        self,
        parent: QtWidgets.QWidget,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> QtWidgets.QWidget:
        editor = super().createEditor(parent, option, index)
        if self.validator:
            editor.setValidator(self.validator)
        return editor

    def setEditorData(
        self, editor: QtWidgets.QWidget, index: QtCore.QModelIndex
    ) -> None:
        super().setEditorData(editor, index)

        # selectAll gets called after opening the editor
        def set_selection():
            editor.selectionChanged.disconnect(set_selection)
            text = editor.text()
            root, ext = os.path.splitext(text)
            editor.setSelection(0, len(root))

        editor.selectionChanged.connect(set_selection)


class FileBrowser(ElementBrowser):
    dir_name = 'Unnamed'
    file_name = 'Unnamed'
    file_filter = ''
    sync_files = True

    def __init__(
        self,
        path: str,
        fields: list[Field] | None = None,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(fields, parent)

        self.path = os.path.normpath(path)

        self._init_elements()

        self.model.element_moved.connect(self._move_element)
        self.model.element_changed.connect(self._change_element)

    def _init_ui(self):
        super()._init_ui()

        delegate = FileNameDelegate(parent=self)
        self.tree.setItemDelegateForColumn(0, delegate)

        icon = MaterialIcon('refresh')
        action = QtWidgets.QAction(icon, 'Refresh', self)
        action.triggered.connect(self.refresh)
        self.toolbar.addAction(action)

    def _init_elements(self) -> None:
        for root, dirs, files in os.walk(self.path):
            indexes = self.model.find_indexes(root, Field('path'))
            if os.path.normpath(root) == self.path:
                parent = QtCore.QModelIndex()
            elif indexes:
                parent = indexes[0]
            else:
                break

            for name in dirs:
                path = os.path.join(root, name)
                self._append_dir(path, parent)
            for name in files:
                if not name.endswith(self.file_filter):
                    continue
                path = os.path.join(root, name)
                self._append_file(path, parent)

        self.proxy.sort(0)
        self.tree.resize_columns()

    def add_element(self):
        parent = self._current_parent()

        parent_element = self.model.element(parent)
        if isinstance(parent_element, FileElement):
            parent_path = parent_element.path
        else:
            parent_path = self.path

        path = os.path.join(parent_path, self.file_name)
        path = unique_path(path)

        try:
            if not os.path.isdir(os.path.dirname(path)):
                os.makedirs(path)
            f = open(path, 'x')
            f.close()
        except OSError as e:
            logger.exception(e)
            return QtCore.QModelIndex()
        index = self._append_file(path, parent)
        return index

    def add_group(self) -> QtCore.QModelIndex:
        parent = self._current_parent()

        parent_element = self.model.element(parent)
        if isinstance(parent_element, FileElement):
            parent_path = parent_element.path
        else:
            parent_path = self.path

        path = os.path.join(parent_path, self.dir_name)
        path = unique_path(path)

        try:
            os.makedirs(path)
        except OSError as e:
            logger.exception(e)
            return QtCore.QModelIndex()
        index = self._append_dir(path, parent)
        return index

    def duplicate_selected(self) -> list[QtCore.QModelIndex]:
        indexes = []
        for index in self.tree.selected_indexes:
            element = self.model.element(index)
            path = unique_path(element.path)
            try:
                shutil.copy(element.path, path)
            except OSError as e:
                logger.exception(e)
                continue
            copied_index = self.model.duplicate_index(index)
            copied_element = self.model.element(copied_index)
            copied_element.path = path
            self.model.refresh_index(copied_index)
            indexes.append(copied_index)
        return indexes

    def refresh(self) -> None:
        self.blockSignals(True)
        elements = self.selected_elements()

        self.model.clear()

        self.sync_files = False
        self._init_elements()
        self.sync_files = True

        self.select_elements(elements)
        self.blockSignals(False)

    def remove_selected(self) -> None:
        # text = 'Are you sure you want to permanently remove {} and all {} contents?'
        # if len(indices) == 1:
        #     index = self.proxy.mapToSource(indices[0])
        #     item = self.model.itemFromIndex(index)
        #     text = text.format(item.text(), 'its')
        # else:
        #     text = text.format('the selected elements', 'their')
        # result = QtWidgets.QMessageBox.question(self, 'Delete', text)
        # if result == QtWidgets.QMessageBox.StandardButton.No:
        #     return

        try:
            for element in self.selected_elements():
                if os.path.isdir(element.path):
                    shutil.rmtree(element.path)
                elif os.path.isfile(element.path):
                    os.remove(element.path)
        except OSError as e:
            logger.exception(e)
            return
        super().remove_selected()

    def _append_dir(self, path: str, parent: QtCore.QModelIndex) -> QtCore.QModelIndex:
        name = os.path.basename(path)
        element = FileElement(name=name, path=path)
        icon = MaterialIcon('folder')
        index = self.model.append_element(element, icon=icon, parent=parent)
        return index

    def _append_file(self, path: str, parent: QtCore.QModelIndex) -> QtCore.QModelIndex:
        name = os.path.basename(path)
        element = FileElement(name=name, path=path)
        index = self.model.append_element(element, no_children=True, parent=parent)
        return index

    def _change_element(self, element: FileElement, previous: FileElement) -> None:
        if element.name == previous.name:
            return
        source_path = previous.path
        parent_path = os.path.dirname(previous.path)
        destination_path = unique_path(os.path.join(parent_path, element.name))
        try:
            shutil.move(source_path, destination_path)
        except OSError as e:
            logger.exception(e)
        self.refresh()

    def _move_element(self, element: FileElement, parent: QtCore.QModelIndex) -> None:
        parent_element = self.model.element(parent)
        if isinstance(parent_element, FileElement):
            parent_path = parent_element.path
        else:
            parent_path = self.path

        destination_path = os.path.join(parent_path, os.path.basename(element.path))
        source_path = element.path
        if destination_path == source_path:
            return

        try:
            if os.path.exists(destination_path):
                raise FileExistsError(f'File Exists: {destination_path}')
            shutil.move(source_path, destination_path)
        except OSError as e:
            logger.exception(e)
        self.refresh()



from __future__ import annotations

from PySide2 import QtWidgets, QtGui, QtCore

from qt_extensions import theme
from qt_extensions.icons import MaterialIcon


class BaseButton(QtWidgets.QPushButton):
    def __init__(
        self,
        text: str = '',
        icon: QtGui.QIcon | None = None,
        color: QtGui.QColor | str | None = None,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(text=text, parent=parent)

        if icon:
            self.setIcon(icon)

        self._icon_size = self.style().pixelMetric(QtWidgets.QStyle.PM_ButtonIconSize)
        self._icon_size = self.iconSize().width()
        self._icon_off = None
        self._icon_on = None
        self._contents_margins = QtCore.QSize(0, 0)
        self._palette_off = None
        self._palette_on = None

        self.set_color(color)

        self.toggled.connect(self._checked_change)

    def sizeHint(self) -> QtCore.QSize:
        size_hint = super().sizeHint()
        size_hint += self._contents_margins
        return size_hint

    def setIcon(self, icon: QtGui.QIcon, on: bool = False) -> None:
        if on:
            if self._palette_on and isinstance(icon, MaterialIcon):
                icon.set_color(self._palette_on.color(QtGui.QPalette.ButtonText))

                pixmap = icon.pixmap(0)

            self._icon_on = icon
        else:
            self._icon_off = icon
            super().setIcon(icon)

    def set_color(self, color: QtGui.QColor | str | None) -> None:
        palette = self.palette()
        self._palette_off = self.palette()

        if isinstance(color, str):
            color = theme.Color(color)

        if color is None:
            button_color = QtGui.QPalette().color(QtGui.QPalette.Button)
            text_color = QtGui.QPalette().color(QtGui.QPalette.ButtonText)
        else:
            button_color = color.darker(125)
            text_color = self.palette().color(QtGui.QPalette.ButtonText)
            if text_color.valueF() > button_color.valueF() * 0.5:
                text_color = text_color.lighter(150)
            else:
                text_color = text_color.darker(150)

        palette.setColor(QtGui.QPalette.Button, button_color)
        palette.setColor(
            QtGui.QPalette.Disabled, QtGui.QPalette.Button, button_color.darker(150)
        )

        palette.setColor(QtGui.QPalette.Normal, QtGui.QPalette.ButtonText, text_color)
        self.setPalette(palette)
        self._palette_on = palette

    def set_margins(self, w: float | None = None, h: float | None = None) -> None:
        if w is not None:
            self._contents_margins.setWidth(w * self._icon_size)
        if h is not None:
            self._contents_margins.setHeight(h * self._icon_size / 2)

    def _checked_change(self, checked: bool) -> None:
        # BUG: fusion style does not recognize On/Off for QIcons
        # https://bugreports.qt.io/browse/QTBUG-82110
        icon = self._icon_on if checked else self._icon_off
        if icon is not None:
            super().setIcon(icon)


class Button(BaseButton):
    def __init__(
        self,
        text: str = '',
        icon: QtGui.QIcon | None = None,
        color: QtGui.QColor | str | None = None,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(text, icon, color, parent)

        self.set_margins(w=2)
        self.setMinimumHeight(self._icon_size * 2)


class CheckBoxButton(BaseButton):
    def __init__(
        self,
        text: str = '',
        icon: QtGui.QIcon | None = None,
        color: QtGui.QColor | str | None = None,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(text, icon, color, parent)

        self.setCheckable(True)
        self.set_margins(w=0.5)
        # self.setMinimumHeight(self._icon_size * 2)

        style = MaterialIcon.Style.FILLED
        size = self._icon_size

        self.setIcon(MaterialIcon('check_box_outline_blank', style=style, size=size))
        self.setIcon(MaterialIcon('check_box', style=style, size=size), True)

    def _checked_change(self, checked: bool) -> None:
        super()._checked_change(checked)
        self._update_color()

    def _update_color(self) -> None:
        if self._palette_on is None:
            return
        if self.isChecked():
            self.setPalette(self._palette_on)
        else:
            self.setPalette(self._palette_off)

    def set_color(self, color: QtGui.QColor | str | None) -> None:
        super().set_color(color)
        self._update_color()



from PySide2 import QtWidgets, QtGui


class QComboBox(QtWidgets.QComboBox):
    # https://bugreports.qt.io/browse/QTBUG-90595
    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        # https://code.qt.io/cgit/qt/qtbase.git/tree/src/widgets/widgets/qcombobox.cpp#n2975
        painter = QtWidgets.QStylePainter(self)
        painter.setPen(self.palette().color(QtGui.QPalette.Text))

        # draw the combobox frame, focusrect and selected etc.
        opt = QtWidgets.QStyleOptionComboBox()
        self.initStyleOption(opt)
        painter.drawComplexControl(QtWidgets.QStyle.CC_ComboBox, opt)

        if self.currentIndex() < 0:
            opt.palette.setBrush(
                QtGui.QPalette.ButtonText,
                opt.palette.brush(QtGui.QPalette.ButtonText).color(),
            )
            if self.placeholderText():
                opt.currentText = self.placeholderText()

        # draw the icon and text
        painter.drawControl(QtWidgets.QStyle.CE_ComboBoxLabel, opt)



from __future__ import annotations

import dataclasses
import sys
from PySide2 import QtWidgets, QtGui
from PySide2.QtGui import QPalette, QColor

from qt_extensions.typeutils import deep_field


@dataclasses.dataclass
class ColorRoles:
    # https://m2.material.io/design/color/the-color-system.html
    critical: QtGui.QColor = deep_field(QtGui.QColor(236, 64, 122))
    error: QtGui.QColor = deep_field(QtGui.QColor(239, 83, 80))
    warning: QtGui.QColor = deep_field(QtGui.QColor(255, 167, 38))
    info: QtGui.QColor = deep_field(QtGui.QColor(66, 165, 245))
    success: QtGui.QColor = deep_field(QtGui.QColor(102, 187, 106))


@dataclasses.dataclass
class ColorScheme:
    base_bg: QtGui.QColor = deep_field(QtGui.QColor(255, 255, 255))
    base_mg: QtGui.QColor = deep_field(QtGui.QColor(240, 240, 240))
    base_fg: QtGui.QColor = deep_field(QtGui.QColor(245, 245, 245))

    text: QtGui.QColor = deep_field(QtGui.QColor(0, 0, 0))
    primary: QtGui.QColor = deep_field(QtGui.QColor(0, 120, 215))

    black: QtGui.QColor = deep_field(QtGui.QColor(20, 20, 20))


class Color(QtGui.QColor):
    def __new__(cls, role: str):
        if role == 'primary':
            return QtGui.QPalette().color(QtGui.QPalette.Highlight)

        roles = ColorRoles()
        try:
            return getattr(roles, role)
        except AttributeError:
            return QtGui.QColor(role)


def apply_scheme(scheme: ColorScheme, palette: QtGui.QPalette):
    # base
    palette.setColor(QtGui.QPalette.Window, scheme.base_mg)
    palette.setColor(QtGui.QPalette.WindowText, scheme.text)
    palette.setColor(QtGui.QPalette.Base, scheme.base_bg)
    palette.setColor(QtGui.QPalette.AlternateBase, scheme.base_fg)
    palette.setColor(QtGui.QPalette.ToolTipBase, scheme.base_bg)
    palette.setColor(QtGui.QPalette.ToolTipText, scheme.text)
    palette.setColor(QtGui.QPalette.PlaceholderText, scheme.text)
    palette.setColor(QtGui.QPalette.Text, scheme.text)
    palette.setColor(QtGui.QPalette.Button, scheme.base_mg)
    palette.setColor(QtGui.QPalette.ButtonText, scheme.text)
    palette.setColor(QtGui.QPalette.BrightText, invert_value(scheme.text))

    palette.setColor(QtGui.QPalette.Highlight, scheme.primary)
    palette.setColor(QtGui.QPalette.HighlightedText, QtGui.QColor(255, 255, 255))

    palette.setColor(QtGui.QPalette.Link, scheme.primary.darker(125))
    palette.setColor(QtGui.QPalette.LinkVisited, scheme.primary.darker(125))

    # auto generate Light, MidLight, Mid, Dark, Shadow colors
    h, s, button_v, a = scheme.base_mg.getHsvF()
    light_v = scheme.base_bg.valueF()
    if scheme.text.value() < scheme.base_mg.value():
        #  text < shadow < dark < mid < button < midlight < light
        black_v = scheme.black.valueF()
    else:
        # shadow < light < midlight < button < mid < dark < text
        black_v = scheme.text.valueF()

    dark = QtGui.QColor.fromHsvF(h, s, lerp(black_v, button_v, 0.35), a)
    mid = QtGui.QColor.fromHsvF(h, s, lerp(black_v, button_v, 0.65), a)
    mid_light = QtGui.QColor.fromHsvF(h, s, lerp(button_v, light_v, 0.5), a)

    palette.setColor(QtGui.QPalette.Light, scheme.base_bg)
    palette.setColor(QtGui.QPalette.Midlight, mid_light)
    palette.setColor(QtGui.QPalette.Mid, mid)
    palette.setColor(QtGui.QPalette.Dark, dark)
    palette.setColor(QtGui.QPalette.Shadow, scheme.black)

    # disabled
    palette.setColor(QPalette.Disabled, QtGui.QPalette.WindowText, dark)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.Base, scheme.base_mg)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.AlternateBase, scheme.base_mg)

    palette.setColor(QPalette.Disabled, QtGui.QPalette.PlaceholderText, dark)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.Text, dark)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.Button, scheme.base_fg)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.ButtonText, dark)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.BrightText, scheme.base_bg)

    palette.setColor(QPalette.Disabled, QPalette.Highlight, scheme.base_mg)
    palette.setColor(QPalette.Disabled, QPalette.HighlightedText, scheme.base_fg)

    palette.setColor(QPalette.Disabled, QtGui.QPalette.Link, dark)
    palette.setColor(QPalette.Disabled, QtGui.QPalette.LinkVisited, dark)


def apply_theme(
    scheme: ColorScheme | None = None,
    style: str = 'Fusion',
    app: QtWidgets.QApplication | None = None,
) -> None:
    if app is None:
        app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    palette = app.palette()
    app.setStyle(style)

    if not scheme:
        return
    apply_scheme(scheme, palette)
    app.setPalette(palette)

    if scheme.text.value() > scheme.base_mg.value():
        # QAbstractItemView alternating row color fix
        item_view_palette = QPalette(palette)
        item_view_palette.setColor(
            QPalette.AlternateBase, palette.color(QPalette.Window)
        )
        app.setPalette(item_view_palette, 'QAbstractItemView')


def invert_value(color: QtGui.QColor) -> QtGui.QColor:
    h, s, v, a = color.getHsvF()
    color = QtGui.QColor.fromHsvF(h, s, 1 - v, a)
    return color


def lerp(a: float, b: float, t: float) -> float:
    return (1 - t) * a + t * b


modern_dark = ColorScheme(
    base_bg=QtGui.QColor(42, 42, 42),
    base_mg=QtGui.QColor(53, 53, 53),
    base_fg=QtGui.QColor(66, 66, 66),
    text=QtGui.QColor(200, 200, 200),
    primary=QColor(42, 165, 218),
)

modern_light = ColorScheme(
    base_bg=QtGui.QColor(255, 255, 255),
    base_mg=QtGui.QColor(242, 242, 242),
    base_fg=QtGui.QColor(230, 235, 240),
    text=QtGui.QColor(0, 0, 0),
    primary=QtGui.QColor(51, 105, 214),
)

monokai = ColorScheme(
    base_bg=QtGui.QColor(41, 42, 46),
    base_mg=QtGui.QColor(50, 51, 56),
    base_fg=QtGui.QColor(60, 61, 66),
    text=QColor(206, 209, 217),
    primary=QtGui.QColor(73, 174, 238),
)



from __future__ import annotations

from enum import auto, Enum

from PySide2 import QtWidgets, QtCore, QtGui


class FlexItemDelegate(QtWidgets.QItemDelegate):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.text_margins = QtCore.QMargins(4, 4, 4, 4)

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ) -> None:
        # draw frame
        self._draw_frame(painter, option)

        # draw decoration and display
        option_view_item = QtWidgets.QStyleOptionViewItem(option)
        # option_view_item.state &= ~QtWidgets.QStyle.State_Selected

        super().paint(painter, option_view_item, index)

        # draw selection
        # self._draw_selection(painter, option)

        # draw focus
        super().drawFocus(painter, option, option.rect)

    def drawFocus(self, painter, option, rect):
        return

    def drawDecoration(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        rect: QtCore.QRect,
        pixmap: QtGui.QPixmap,
    ) -> None:
        option.rect.adjust(2, 2, -2, -2)
        # option.rect.adjust(self.text_margins.left(), self.text_margins.top(), -self.text_margins.right(), 0)

        font_metrics = painter.fontMetrics()
        height = font_metrics.height()
        height += self.text_margins.top() + self.text_margins.bottom()

        rect = QtCore.QRect(option.rect)
        rect.setHeight(option.rect.height() - height)

        if pixmap is None or not rect.isValid():
            return

        pixmap = pixmap.scaled(rect.size(), QtCore.Qt.KeepAspectRatioByExpanding)

        # if hasattr(painter, 'selected') and painter.selected:
        if option.state & QtWidgets.QStyle.State_Selected:
            pixmap = self.selectedPixmap(
                pixmap, option.palette, option.state & QtWidgets.QStyle.State_Enabled
            )

        source_rect = QtWidgets.QStyle.alignedRect(
            option.direction, option.decorationAlignment, rect.size(), pixmap.rect()
        )

        painter.drawPixmap(rect, pixmap, source_rect)

    def drawDisplay(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        rect: QtCore.QRect,
        text: str,
    ) -> None:
        rect.adjust(
            self.text_margins.left(),
            0,
            -self.text_margins.right(),
            -self.text_margins.bottom(),
        )
        option.state &= ~QtWidgets.QStyle.State_Selected
        super().drawDisplay(painter, option, rect, text)

    def sizeHint(
        self, option: QtWidgets.QStyleOptionViewItem, index: QtCore.QModelIndex
    ) -> QtCore.QSize:
        size_hint = super().sizeHint(option, index)
        # size_hint = option.rect.size()
        return size_hint

        # value = index.data(QtCore.Qt.SizeHintRole)
        # if value.isValid():
        #     return value
        #
        # decorationRect = self.rect(option, index, QtCore.Qt.DecorationRole)
        # checkRect = self.rect(option, index, QtCore.Qt.CheckStateRole)
        # displayRect = self.displayRect(index, option, decorationRect, checkRect)
        #
        # # doLayout(option, &checkRect, &decorationRect, &displayRect, True)
        # return (decorationRect | displayRect | checkRect).size()

    @staticmethod
    def _draw_frame(
        painter: QtGui.QPainter, option: QtWidgets.QStyleOptionViewItem
    ) -> None:
        widget = option.widget
        style = widget.style() or QtWidgets.QApplication.style()
        option_frame = QtWidgets.QStyleOptionFrame()

        palette = option_frame.palette
        option_frame.palette.setColor(
            QtGui.QPalette.Window, palette.color(QtGui.QPalette.Dark)
        )
        option_frame.state = option.state
        option_frame.rect = option.rect
        option_frame.rect.adjust(1, 1, -1, -1)
        option_frame.frameShape = QtWidgets.QFrame.StyledPanel
        option_frame.lineWidth = 1

        if option.state & QtWidgets.QStyle.State_Selected:
            color = palette.color(QtGui.QPalette.Highlight)
        else:
            color = palette.color(QtGui.QPalette.Window)
        painter.fillRect(option_frame.rect, color)
        painter.save()
        painter.setRenderHint(QtGui.QPainter.Antialiasing, False)
        style.drawPrimitive(QtWidgets.QStyle.PE_Frame, option_frame, painter)
        # painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        painter.restore()

    @staticmethod
    def _draw_selection(
        painter: QtGui.QPainter, option: QtWidgets.QStyleOptionViewItem
    ) -> None:
        if option.state & QtWidgets.QStyle.State_Selected:
            palette = option.palette
            color = palette.color(QtGui.QPalette.Highlight)
            pen = QtGui.QPen(color)
            painter.save()
            painter.setPen(pen)
            painter.drawRect(option.rect)
            painter.restore()


class FlexView(QtWidgets.QAbstractItemView):
    class PositionFlags(Enum):
        START = auto()
        END = auto()
        CENTER = auto()
        STRETCH = auto()

        SPACE_BETWEEN = auto()
        # SPACE_AROUND = auto()
        # SPACE_EVENLY = auto()

    class LayoutFlags(Enum):
        ROW = auto()
        # ROW_REVERSE = auto()
        # COLUMN = auto()
        # COLUMN_REVERSE = auto()

    class WrapFlags(Enum):
        NONE = 0
        WRAP = auto()
        # WRAP_REVERSE = auto()

    flex_direction: LayoutFlags = LayoutFlags.ROW
    align_content: PositionFlags = PositionFlags.START
    justify_content: PositionFlags = PositionFlags.START
    align_items: PositionFlags = PositionFlags.STRETCH
    wrap: WrapFlags = WrapFlags.WRAP
    grow: bool = False

    min_size = QtCore.QSize(0, 0)
    default_size = QtCore.QSize(250, 150)
    child_rows = True
    column = 0

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._item_rects = []
        self._rubber_origin = QtCore.QPoint()
        self._rubber_band = None

        self.height = 0
        self.min_item_width = 0

        self.contents_margins = QtCore.QMargins(6, 6, 6, 6)
        self.spacing = 4

        # defaults
        self.setItemDelegate(FlexItemDelegate())
        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)

        self.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

    @property
    def item_rects(self) -> dict[QtCore.QModelIndex, QtCore.QRect]:
        if not self._item_rects:
            self._item_rects = self._update_item_rects()
        return self._item_rects

    @item_rects.setter
    def item_rects(self, value: dict[QtCore.QModelIndex, QtCore.QRect]) -> None:
        self._item_rects = value

    def dataChanged(
        self,
        top_left: QtCore.QModelIndex,
        bottom_right: QtCore.QModelIndex,
        roles: list[QtCore.Qt.ItemDataRole] | None = None,
    ) -> None:
        if roles is None:
            roles = []
        self.item_rects = {}
        super().dataChanged(top_left, bottom_right, roles)

    def indexAt(self, point: QtCore.QPoint) -> QtCore.QModelIndex:
        offset = QtCore.QPoint(
            self.horizontalScrollBar().value(), self.verticalScrollBar().value()
        )
        point += offset

        self.item_rects = {}

        for index, rect in self.item_rects.items():
            if rect.contains(point):
                return index
        return self.rootIndex()

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mousePressEvent(event)

        self.indexAt(event.pos())

        # rubber_band
        if event.button() == QtCore.Qt.LeftButton:
            if not self._rubber_band:
                self._rubber_band = QtWidgets.QRubberBand(
                    QtWidgets.QRubberBand.Rectangle, self
                )
            self._rubber_origin = event.pos()
            rect = QtCore.QRect(self._rubber_origin, QtCore.QSize())
            self._rubber_band.setGeometry(rect)
            self._rubber_band.show()

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseMoveEvent(event)

        if isinstance(self._rubber_band, QtWidgets.QWidget):
            rect = QtCore.QRect(self._rubber_origin, event.pos())
            self._rubber_band.setGeometry(rect.normalized())

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        if isinstance(self._rubber_band, QtWidgets.QWidget):
            self._rubber_band.hide()

    def moveCursor(
        self,
        cursor_action: QtWidgets.QAbstractItemView.CursorAction,
        modifiers: QtCore.Qt.KeyboardModifiers,
    ) -> QtCore.QModelIndex:
        view = QtWidgets.QAbstractItemView
        index = self.currentIndex()
        if index.isValid():
            row = index.row()

            min_row = 0
            max_row = self.model().rowCount(index.parent()) - 1

            if cursor_action in (view.MoveLeft | view.MoveUp | view.MovePrevious):
                row -= 1
            elif cursor_action in (view.MoveRight | view.MoveDown | view.MoveNext):
                row += 1
            elif cursor_action == view.MoveHome:
                row = min_row
            elif cursor_action == view.MoveEnd:
                row = max_row
            elif cursor_action == view.MovePageUp:
                row -= 10
            elif cursor_action == view.MovePageDown:
                row += 10
            else:
                return index

            row = max(min_row, min(max_row, row))

            index = self.model().index(row, index.column(), index.parent())
        return index

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self.viewport())

        painter.setRenderHints(
            QtGui.QPainter.Antialiasing | QtGui.QPainter.TextAntialiasing
        )

        option = QtWidgets.QStyleOptionViewItem(self.viewOptions())
        state = option.state
        enabled = bool(state & QtWidgets.QStyle.State_Enabled)
        has_focus = self.hasFocus() or self.viewport().hasFocus()
        focused = has_focus and self.currentIndex().isValid()

        for index in self._indexes():
            rect = self.item_rects.get(index)
            if rect is None:
                continue
            option.state = state
            option.rect = self._map_to_viewport(rect)
            option.decorationAlignment = QtCore.Qt.AlignCenter
            option.decorationPosition = QtWidgets.QStyleOptionViewItem.Top
            option.displayAlignment = QtCore.Qt.AlignBottom | QtCore.Qt.AlignLeft

            if self.selectionModel().isSelected(index):
                option.state |= QtWidgets.QStyle.State_Selected
            if enabled:
                if self.model().flags(index) & QtCore.Qt.ItemIsEnabled:
                    current_color_group = QtGui.QPalette.Normal
                else:
                    option.state &= ~QtWidgets.QStyle.State_Enabled
                    current_color_group = QtGui.QPalette.Disabled
                option.palette.setCurrentColorGroup(current_color_group)
            if focused and self.currentIndex() == index:
                option.state |= QtWidgets.QStyle.State_HasFocus
                if self.state() == QtWidgets.QAbstractItemView.EditingState:
                    option.state |= QtWidgets.QStyle.State_Editing
            self.itemDelegate(index).paint(painter, option, index)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        self.item_rects = {}
        super().resizeEvent(event)

    def rowsAboutToBeRemoved(
        self, parent: QtCore.QModelIndex, start: int, end: int
    ) -> None:
        self.item_rects = {}
        super().rowsAboutToBeRemoved(parent, start, end)

    def rowsInserted(self, parent: QtCore.QModelIndex, start: int, end: int) -> None:
        self.item_rects = {}
        super().rowsInserted(parent, start, end)

    def setModel(self, model: QtCore.QAbstractItemModel) -> None:
        self.item_rects = {}
        super().setModel(model)

    def scrollTo(
        self,
        index: QtCore.QModelIndex,
        hint: QtWidgets.QAbstractItemView.ScrollHint = QtWidgets.QAbstractItemView.EnsureVisible,
    ) -> None:
        if index.parent() != self.rootIndex():
            return

        if not index.isValid():
            return

        rect = self.item_rects[index]
        mapped_rect = self._map_to_viewport(rect)

        if (
            hint == QtWidgets.QAbstractItemView.EnsureVisible
            and self.viewport().rect().contains(mapped_rect)
        ):
            self.viewport().update(mapped_rect)
            return

        horizontal_value = self._horizontal_scroll_to_value(rect, hint)
        self.horizontalScrollBar().setValue(horizontal_value)
        vertical_value = self._vertical_scroll_to_value(rect, hint)
        self.verticalScrollBar().setValue(vertical_value)

    def horizontalOffset(self) -> int:
        return self.horizontalScrollBar().value()

    def verticalOffset(self) -> int:
        return self.verticalScrollBar().value()

    def selectionChanged(
        self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection
    ) -> None:
        super().selectionChanged(selected, deselected)
        self.viewport().update()

    def setSelection(
        self, rect: QtCore.QRect, command: QtCore.QItemSelectionModel.SelectionFlags
    ) -> None:
        rect = rect.translated(self.horizontalOffset(), self.verticalOffset())
        rect = rect.normalized()

        update_rect = QtCore.QRect()
        selection = QtCore.QItemSelection()
        for index, item_rect in self.item_rects.items():
            if item_rect.intersects(rect):
                update_rect = update_rect.united(item_rect)
                selection.select(index, index)
        self.selectionModel().select(selection, command)

    def update(self) -> None:
        super().update()
        self.viewport().update()

    def updateGeometries(self) -> None:
        self._update_scrollbars()
        super().updateGeometries()

    def visualRect(self, index: QtCore.QModelIndex) -> QtCore.QRect:
        if index.isValid():
            rect = self.item_rects[index]
        else:
            rect = QtCore.QRect()
        return self._map_to_viewport(rect)

    def visualRegionForSelection(
        self, selection: QtCore.QItemSelection
    ) -> QtGui.QRegion:
        region = QtGui.QRegion()

        for index in selection.indexes():
            rect = self.item_rects[index]
            region = region.united(self._map_to_viewport(rect))

        return region

    def _indexes(
        self, parent: QtCore.QModelIndex | None = None
    ) -> list[QtCore.QModelIndex]:
        if parent is None:
            parent = self.rootIndex()
        indexes = []
        for row in range(self.model().rowCount(parent)):
            index = self.model().index(row, self.column, parent)
            if index.data():  # TODO: temporary
                indexes.append(index)
            if self.child_rows:
                first_index = index.siblingAtColumn(0)
                if first_index.isValid():
                    indexes.extend(self._indexes(first_index))
        return indexes

    def _horizontal_scroll_to_value(
        self, rect: QtCore.QRect, hint: QtWidgets.QAbstractItemView.ScrollHint
    ) -> float:
        if hint == QtWidgets.QAbstractItemView.PositionAtBottom:
            value = rect.right() - self.viewport().width() + self.spacing
        elif hint == QtWidgets.QAbstractItemView.PositionAtCenter:
            value = rect.left() + rect.width() / 2 - (self.viewport().width() / 2)
        else:
            value = rect.left() - self.spacing
        return value

    def _vertical_scroll_to_value(
        self, rect: QtCore.QRect, hint: QtWidgets.QAbstractItemView.ScrollHint
    ) -> float:
        if hint == QtWidgets.QAbstractItemView.PositionAtBottom:
            value = rect.bottom() - self.viewport().height() + self.spacing
        elif hint == QtWidgets.QAbstractItemView.PositionAtCenter:
            value = rect.top() + rect.height() / 2 - (self.viewport().height() / 2)
        else:
            value = rect.top() - self.spacing
        return value

    def _map_to_viewport(
        self, rect: QtCore.QRect, extend: bool = False
    ) -> QtCore.QRect:
        if not rect.isValid():
            return rect
        result = QtCore.QRect(rect)
        if extend:
            result.setLeft(self.spacing)
            result.setWidth(
                max(
                    rect.width(),
                    max(self.contentsSize().width(), self.viewport().width())
                    - 2 * self.spacing,
                )
            )

        dx = -self.horizontalOffset()
        dy = -self.verticalOffset()
        rect = result.adjusted(dx, dy, dx, dy)
        return rect

    def _update_item_rects(self) -> dict[QtCore.QModelIndex, QtCore.QRect]:
        if not self.model():
            return {}

        position_flags = self.__class__.PositionFlags
        wrap_flags = self.__class__.WrapFlags

        default_size = self.default_size
        rect = self.viewport().rect()
        rect = rect.marginsRemoved(self.contents_margins)
        x = rect.x()
        y = rect.y()
        max_height = 0
        group_width = 0
        spacing = self.spacing
        item_rects = {}
        min_item_width = 0
        previous_item_width = default_size.width()

        indexes = self._indexes()
        group_indexes = []
        for i, index in enumerate(indexes):
            group_indexes.append(index)
            max_height = max(max_height, default_size.height())
            next_x = x + default_size.width() + spacing
            group_width += default_size.width() + spacing
            x = next_x

            try:
                # next_next_x = next_x + items[i + 1].sizeHint().width()
                next_next_x = next_x + default_size.width()
            except IndexError:
                next_next_x = next_x - spacing

            is_last_item = i == len(indexes) - 1
            is_overlapped = next_next_x > rect.right()

            if self.wrap == wrap_flags.WRAP:
                end_group = is_overlapped or is_last_item
            else:
                end_group = True

            if end_group:
                count = len(group_indexes)
                item_spacing = spacing
                group_width -= spacing

                # x position
                if self.justify_content == position_flags.START or self.grow:
                    item_x = rect.x()
                elif self.justify_content == position_flags.END:
                    item_x = rect.right() - group_width
                elif self.justify_content == position_flags.CENTER:
                    total_space = rect.width() - group_width
                    item_x = rect.x() + total_space / 2
                elif self.justify_content == position_flags.SPACE_BETWEEN:
                    total_space = rect.width() - group_width + (count - 1) * spacing
                    item_x = rect.x()
                    item_spacing = total_space / max(1, count - 1)
                else:
                    item_x = rect.x()

                item_width = (rect.width() - (count - 1) * spacing) / count
                for group_index in group_indexes:
                    # size = group_item.sizeHint()
                    size = default_size
                    item_height = size.height()

                    # y position
                    if self.align_items == position_flags.START:
                        item_y = y
                    elif self.align_items == position_flags.END:
                        item_y = y + max_height - size.height()
                    elif self.align_items == position_flags.CENTER:
                        item_y = y + (max_height - size.height()) / 2
                    elif self.align_items == position_flags.STRETCH:
                        item_y = y
                        item_height = max_height
                    else:
                        item_y = y

                    # size
                    if self.grow:
                        if is_last_item:
                            item_width = previous_item_width
                        else:
                            previous_item_width = item_width
                    else:
                        item_width = size.width()

                    item_rect = QtCore.QRect(
                        QtCore.QPoint(item_x, item_y),
                        QtCore.QSize(item_width, item_height),
                    )

                    min_item_width = max(item_rect.width(), min_item_width)
                    item_rects[group_index] = item_rect
                    item_x += item_width + item_spacing

                x = rect.x()
                y = y + max_height + spacing

                max_height = 0
                group_indexes = []
                group_width = 0

        rect.setBottom(y - spacing)
        rect = rect.marginsAdded(self.contents_margins)
        self.height = rect.height()
        self.min_item_width = (
            min_item_width
            + self.contents_margins.left()
            + self.contents_margins.right()
        )

        return item_rects

    def _update_scrollbars(self) -> None:
        self._update_item_rects()

        viewport = self.viewport()

        self.verticalScrollBar().setPageStep(viewport.height())
        self.verticalScrollBar().setRange(0, self.height - viewport.height())

        self.horizontalScrollBar().setPageStep(viewport.width())
        self.horizontalScrollBar().setRange(0, self.min_item_width - viewport.width())



from __future__ import annotations

import math
from enum import Enum, IntEnum, auto, EnumMeta
from functools import partial
from typing import Any, Callable, Optional

from PySide2 import QtWidgets, QtGui, QtCore

from qt_extensions import helper
from qt_extensions.button import BaseButton
from qt_extensions.icons import MaterialIcon
from qt_extensions.resizegrip import ResizeGrip


class ParameterWidget(QtWidgets.QWidget):
    # by not using @property it is easier to set typing hints
    enabled_changed: QtCore.Signal = QtCore.Signal(bool)
    value_changed: QtCore.Signal = QtCore.Signal(object)

    _value: Any = None
    _default: Any = None
    _name: str = ''
    _label: str = ''
    _tooltip: str = ''

    def __init__(self, name: str = '', parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        if name:
            self.set_name(name)
            self.set_label(helper.title(name))

        self._init_layout()
        self._init_ui()
        self._init_defaults()

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.name())})'

    def _init_layout(self) -> None:
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

    def _init_ui(self) -> None:
        pass

    def _init_defaults(self) -> None:
        self.blockSignals(True)
        attributes = dir(self.__class__)
        for name in attributes:
            if not name.startswith('_') or name.startswith('__') or name == '_value':
                continue
            try:
                setter = getattr(self, f'set{name}')
                value = getattr(self, name)
            except AttributeError:
                continue
            setter(value)
        self.blockSignals(False)

    def changeEvent(self, event):
        if event.type() == QtCore.QEvent.EnabledChange:
            self.enabled_changed.emit(self.isEnabled())
        super().changeEvent(event)

    def default(self) -> Any:
        return self._default

    def label(self) -> str:
        return self._label

    def name(self) -> str:
        return self._name

    def set_default(self, default: Any) -> None:
        self.set_value(default)
        self._default = self.value()

    def set_label(self, label: str) -> None:
        self._label = label

    def set_name(self, name: str) -> None:
        self._name = name

    def set_tooltip(self, tooltip: str) -> None:
        self._tooltip = tooltip

    def set_value(self, value: Any) -> None:
        if value != self._value:
            self._value = value
            self.value_changed.emit(value)

    def tooltip(self) -> str:
        return self._tooltip

    def value(self) -> Any:
        return self._value


class IntParameter(ParameterWidget):
    value_changed: QtCore.Signal = QtCore.Signal(int)

    _value: int = 0
    _default: int = 0
    _slider_min: int = 0
    _slider_max: int = 10
    _line_min: Optional[int] = None
    _line_max: Optional[int] = None
    _slider_visible: bool = True
    _commit_on_edit: bool = False

    def _init_ui(self) -> None:
        # line
        self.line = IntLineEdit(self)
        self.line.value_changed.connect(self._line_value_changed)
        self.layout().addWidget(self.line)

        # slider
        self.slider = IntSlider()
        self.slider.value_changed.connect(self._slider_value_changed)
        # prevent any size changes when slider shows
        self.slider.setMaximumHeight(self.line.minimumSizeHint().height())
        self.layout().addWidget(self.slider)
        self.layout().setStretch(1, 1)

        self.setFocusProxy(self.line)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self._toggle_slider(True)

    def commit_on_edit(self) -> bool:
        return self._commit_on_edit

    def line_min(self) -> int:
        return self._line_min

    def line_max(self) -> int:
        return self._line_max

    def slider_min(self) -> int:
        return self._slider_min

    def slider_max(self) -> int:
        return self._slider_max

    def slider_visible(self) -> bool:
        return self._slider_visible

    def set_commit_on_edit(self, commit_on_edit: bool) -> None:
        self._commit_on_edit = commit_on_edit
        self.line.commit_on_edit = commit_on_edit

    def set_line_min(self, line_min: Optional[int]) -> None:
        self._line_min = line_min
        self.line.set_minimum(line_min)

    def set_line_max(self, line_max: Optional[int]) -> None:
        self._line_max = line_max
        self.line.set_maximum(line_max)

    def set_slider_min(self, slider_min: int) -> None:
        self._slider_min = slider_min
        self.slider.set_minimum(slider_min)

    def set_slider_max(self, slider_max: int) -> None:
        self._slider_max = slider_max
        self.slider.set_maximum(slider_max)

    def set_slider_visible(self, slider_visible: bool) -> None:
        self._slider_visible = slider_visible
        self._toggle_slider(slider_visible)

    def set_value(self, value: int) -> None:
        super().set_value(value)
        self._set_line_value(value)
        self._set_slider_value(value)

    def value(self) -> int:
        return super().value()

    def _line_value_changed(self, value: int) -> None:
        super().set_value(value)
        self._set_slider_value(value)

    def _slider_value_changed(self, value: int) -> None:
        super().set_value(value)
        self._set_line_value(value)

    def _set_line_value(self, value: int) -> None:
        self.line.blockSignals(True)
        self.line.set_value(value)
        self.line.blockSignals(False)

    def _set_slider_value(self, value: int) -> None:
        self.slider.blockSignals(True)
        self.slider.set_value(value)
        self.slider.blockSignals(False)

    def _toggle_slider(self, value: bool) -> None:
        has_space = self.size().width() > 200
        self.slider.setVisible(self._slider_visible and value and has_space)


class FloatParameter(IntParameter):
    value_changed: QtCore.Signal = QtCore.Signal(float)

    _value: float = 0
    _default: float = 0
    _slider_min: float = 0
    _slider_max: float = 1
    _line_min: Optional[float] = None
    _line_max: Optional[float] = None
    _decimals: int = 4

    def _init_ui(self) -> None:
        # line
        self.line = FloatLineEdit(self)
        self.line.value_changed.connect(self._line_value_changed)
        self.layout().addWidget(self.line)

        # slider
        self.slider = FloatSlider()
        self.slider.value_changed.connect(self._slider_value_changed)
        # prevent any size changes when slider shows
        self.slider.setMaximumHeight(self.line.minimumSizeHint().height())
        self.layout().addWidget(self.slider)
        self.layout().setStretch(1, 1)

        self.setFocusProxy(self.line)

    def decimals(self) -> int:
        return self._decimals

    def line_min(self) -> float:
        return super().line_min()

    def line_max(self) -> float:
        return super().line_max()

    def slider_min(self) -> float:
        return super().slider_min()

    def slider_max(self) -> float:
        return super().slider_max()

    def set_decimals(self, decimals: int) -> None:
        self._decimals = decimals
        self.line.set_decimals(decimals)

    def set_line_min(self, line_min: Optional[float]) -> None:
        super().set_line_min(line_min)

    def set_line_max(self, line_max: Optional[float]) -> None:
        super().set_line_max(line_max)

    def set_slider_min(self, slider_min: float) -> None:
        super().set_slider_min(slider_min)

    def set_slider_max(self, slider_max: float) -> None:
        super().set_slider_max(slider_max)

    def set_value(self, value: float) -> None:
        super().set_value(value)

    def value(self) -> float:
        return super().value()


class StringParameter(ParameterWidget):
    value_changed: QtCore.Signal = QtCore.Signal(str)

    _value: str = ''
    _default: str = ''
    _area: bool = False
    _menu: dict | list | tuple | None = None

    def _init_ui(self) -> None:
        if self._area:
            self.text = TextEdit()
            self.text.editing_finished.connect(self._editing_finished)
            ResizeGrip(self.text)
        else:
            self.text = QtWidgets.QLineEdit()
            self.text.editingFinished.connect(self._editing_finished)
        self.layout().addWidget(self.text)
        self.setFocusProxy(self.text)

        self.menu_button = QtWidgets.QToolButton()
        self.menu_button.setAutoRaise(True)
        self.layout().addWidget(self.menu_button)
        self.menu_button.hide()

    def area(self) -> bool:
        return self._area

    def menu(self) -> dict | list | tuple | None:
        return self._menu

    def set_area(self, area: bool) -> None:
        self._area = area
        for i in reversed(range(self.layout().count())):
            self.layout().itemAt(i).widget().deleteLater()
        self._init_ui()

    def set_menu(self, menu: dict | list | tuple | None) -> None:
        self._menu = menu

        # update menu
        if not self._area and self._menu is not None:
            if not self.menu_button.defaultAction():
                # build dynamically for optimization
                icon = MaterialIcon('expand_more')
                action = QtWidgets.QAction(icon, 'Fill', self)
                action.triggered.connect(self._show_menu)
                self.menu_button.setDefaultAction(action)
            self.menu_button.show()
        else:
            self.menu_button.hide()

    def set_value(self, value: str) -> None:
        super().set_value(value)
        self.text.blockSignals(True)
        if self._area:
            self.text.setPlainText(value)
        else:
            self.text.setText(value)
        self.text.blockSignals(False)

    def value(self) -> str:
        return super().value()

    def _build_menu(
        self, content: dict | list | tuple, menu: QtWidgets.QMenu | None = None
    ) -> QtWidgets.QMenu:
        if menu is None:
            menu = QtWidgets.QMenu(self)
        if isinstance(content, (list, tuple)):
            content = {i: i for i in content}
        for label, text in content.items():
            if isinstance(text, dict):
                sub_menu = menu.addMenu(label)
                self._build_menu(text, sub_menu)
            else:
                action = QtWidgets.QAction(label, self)
                action.triggered.connect(partial(self.set_value, str(text)))
                menu.addAction(action)
        return menu

    def _editing_finished(self) -> None:
        if self._area:
            value = self.text.toPlainText()
        else:
            value = self.text.text()
        super().set_value(value)

    def _show_menu(self) -> None:
        relative_pos = self.menu_button.rect().topRight()
        relative_pos.setX(relative_pos.x() + 2)
        position = self.menu_button.mapToGlobal(relative_pos)

        menu = self._build_menu(self._menu)
        menu.exec_(position)
        self.menu_button.setDown(False)


class PathParameter(ParameterWidget):
    class Method(IntEnum):
        OPEN_FILE = auto()
        SAVE_FILE = auto()
        EXISTING_DIR = auto()

    OPEN_FILE = Method.OPEN_FILE
    SAVE_FILE = Method.SAVE_FILE
    EXISTING_DIR = Method.EXISTING_DIR

    value_changed: QtCore.Signal = QtCore.Signal(str)

    _value: str = ''
    _default: str = ''
    _method: Method = OPEN_FILE
    _dir_fallback: str = ''

    def _init_ui(self) -> None:
        self.line = QtWidgets.QLineEdit()
        self.line.editingFinished.connect(self._editing_finished)
        self.layout().addWidget(self.line)

        self.button = QtWidgets.QToolButton()
        self.button.setIcon(MaterialIcon('file_open'))
        self.button.clicked.connect(self.browse)
        self.layout().addWidget(self.button)

        self.layout().setStretch(0, 1)
        self.setFocusProxy(self.line)

    def browse(self) -> None:
        start_dir = self._value or self._dir_fallback
        if self._method == PathParameter.OPEN_FILE:
            path, filters = QtWidgets.QFileDialog.getOpenFileName(
                parent=self,
                caption='Open File',
                dir=start_dir,
            )
        elif self._method == PathParameter.SAVE_FILE:
            path, filters = QtWidgets.QFileDialog.getSaveFileName(
                parent=self,
                caption='Save File',
                dir=start_dir,
                filter='*.*',
            )
        elif self._method == PathParameter.EXISTING_DIR:
            path = QtWidgets.QFileDialog.getExistingDirectory(
                parent=self,
                caption='Select Directory',
                dir=start_dir,
            )
        else:
            return

        if path:
            self.set_value(path)

    def dir_fallback(self) -> str:
        return self._dir_fallback

    def method(self) -> Method:
        return self._method

    def set_dir_fallback(self, dir_fallback: str) -> None:
        self._dir_fallback = dir_fallback

    def set_method(self, method: Method) -> None:
        self._method = method

    def set_value(self, value: str) -> None:
        super().set_value(value)
        self.line.blockSignals(True)
        self.line.setText(value)
        self.line.blockSignals(False)

    def value(self) -> str:
        return super().value()

    def _editing_finished(self) -> None:
        value = self.line.text()
        super().set_value(value)


class EnumParameter(ParameterWidget):
    _value: Enum = None
    _default: Enum = None
    _formatter: Callable = staticmethod(helper.title)
    _enum: EnumMeta | None = None

    def _init_ui(self) -> None:
        self.combo = QtWidgets.QComboBox()
        self.combo.currentIndexChanged.connect(self._current_index_changed)

        self.layout().addWidget(self.combo)
        self.setFocusProxy(self.combo)

    def enum(self) -> EnumMeta:
        return self._enum

    def formatter(self) -> Callable:
        return self._formatter

    def set_enum(self, enum: EnumMeta) -> None:
        self._enum = enum
        self._update_items()
        self.combo.setCurrentIndex(0)

    def set_formatter(self, formatter: Callable) -> None:
        self._formatter = formatter
        index = self.combo.currentIndex()
        self._update_items()
        self.combo.setCurrentIndex(index)

    def set_value(self, value: Any) -> None:
        value = self._enum_from_value(value)
        super().set_value(value)

        self.combo.blockSignals(True)
        if value is None:
            index = -1
        else:
            index = self.combo.findData(value.value)
        self.combo.setCurrentIndex(index)
        self.combo.blockSignals(False)

    def value(self) -> Enum | None:
        return super().value()

    def _current_index_changed(self, index: int) -> None:
        value = self.combo.itemData(index)
        value = self._enum_from_value(value)
        super().set_value(value)

    def _enum_from_value(self, value: Any) -> Enum | None:
        try:
            # value is Enum
            if isinstance(value, self._enum):
                return value
        except TypeError:
            pass

        try:
            # value is Enum.name
            return self._enum[value]
        except KeyError:
            pass
        except TypeError:
            return None

        try:
            # value is Enum.value
            return self._enum(value)
        except (ValueError, TypeError):
            return None

    def _update_items(self) -> None:
        self.combo.blockSignals(True)
        for index in reversed(range(self.combo.count())):
            self.combo.removeItem(index)

        if isinstance(self._enum, EnumMeta):
            for member in self._enum:
                label = self._formatter(member.name)
                self.combo.addItem(label, member.value)
        self.combo.blockSignals(False)


class BoolParameter(ParameterWidget):
    value_changed: QtCore.Signal = QtCore.Signal(bool)

    _value: bool = False
    _default: bool = False

    def _init_ui(self) -> None:
        self.checkbox = QtWidgets.QCheckBox()
        self.checkbox.toggled.connect(super().set_value)
        self.layout().addWidget(self.checkbox)
        self.layout().addStretch()
        self.setFocusProxy(self.checkbox)

    def set_value(self, value: bool) -> None:
        super().set_value(value)
        self.checkbox.blockSignals(True)
        self.checkbox.setChecked(value)
        self.checkbox.blockSignals(False)

    def value(self) -> bool:
        return super().value()


class MultiIntParameter(IntParameter):
    multi_count = 2
    value_changed: QtCore.Signal = QtCore.Signal(tuple)

    _value: tuple[int, ...] = (0, 0)
    _default: tuple[int, ...] = (0, 0)
    _keep_ratio: bool = True
    _ratio_visible: bool = True

    def _init_ui(self) -> None:
        # lines
        self.lines = []
        for i in range(self.multi_count):
            line = IntLineEdit()
            line.value_changed.connect(self._line_value_changed)
            self.layout().addWidget(line)
            self.lines.append(line)

        # slider
        self.slider = IntSlider()
        self.slider.value_changed.connect(self._slider_value_changed)
        # prevent any size changes when slider shows
        line_height = self.lines[0].minimumSizeHint().height()
        self.slider.setMaximumHeight(line_height)
        self.layout().addWidget(self.slider)
        self.layout().setStretch(self.multi_count, 1)

        # keep ratio button
        self.keep_ratio_button = RatioButton()
        self.keep_ratio_button.setMaximumSize(line_height, line_height)
        self.keep_ratio_button.toggled.connect(self.set_keep_ratio)
        self.layout().addWidget(self.keep_ratio_button)

        self.setFocusProxy(self.lines[0])

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        QtWidgets.QWidget.resizeEvent(self, event)
        if self._keep_ratio:
            self._toggle_slider(True)

    def set_commit_on_edit(self, commit_on_edit: bool) -> None:
        self._commit_on_edit = commit_on_edit
        for line in self.lines:
            line.commit_on_edit = commit_on_edit

    def set_keep_ratio(self, keep_ratio: bool) -> None:
        self._keep_ratio = keep_ratio
        self.keep_ratio_button.setChecked(keep_ratio)
        for line in self.lines[1:]:
            line.setVisible(not keep_ratio)
            if line.value() != self.lines[0].value():
                line.set_value(self.lines[0].value())
        self._toggle_slider(keep_ratio)

    def set_line_min(self, line_min: int) -> None:
        self._line_min = line_min
        for line in self.lines:
            line.set_minimum(line_min)

    def set_line_max(self, line_max: int) -> None:
        self._line_max = line_max
        for line in self.lines:
            line.set_maximum(line_max)

    def set_ratio_visible(self, ratio_visible: bool) -> None:
        self._ratio_visible = ratio_visible
        self.keep_ratio_button.setVisible(ratio_visible)
        for line in self.lines[1:]:
            line.setVisible(not ratio_visible)
        if not ratio_visible:
            self.set_keep_ratio(False)

    def set_value(self, value: list | tuple) -> None:
        if isinstance(value, (list, tuple)):
            values = value
        else:
            values = self._cast_to_tuple(value)
        if isinstance(self, ColorParameter):
            pass
        if not all(values[0] == x for x in values):
            self.set_keep_ratio(False)
        if self._keep_ratio:
            values = (values[0],) * self.multi_count
        ParameterWidget.set_value(self, self._cast_to_type(values))
        self._set_slider_value(values[0])
        self._set_line_values(values)

    def value(self) -> tuple[int, ...]:
        return ParameterWidget.value(self)

    def _line_value_changed(self, value: int) -> None:
        if self._keep_ratio:
            values = (self.lines[0].value(),) * self.multi_count
            for line in self.lines[1:]:
                line.set_value(values[0])
        else:
            values = tuple(line.value() for line in self.lines)

        value = self._cast_to_type(values)
        ParameterWidget.set_value(self, value)
        self._set_slider_value(values[0])

    def _cast_to_tuple(self, values: tuple[int, ...]) -> tuple[int, ...]:
        return values

    def _cast_to_type(self, values: tuple[int, ...]) -> tuple[int, ...]:
        return values

    def _slider_value_changed(self, value: int) -> None:
        values = (value,) * self.multi_count
        value = self._cast_to_type(values)
        ParameterWidget.set_value(self, value)
        self._set_line_values(values)

    def _set_line_values(self, values: tuple[int, ...]) -> None:
        for line, value in zip(self.lines, values):
            line.blockSignals(True)
            line.set_value(value)
            line.blockSignals(False)


class MultiFloatParameter(MultiIntParameter):
    _value: tuple[float, ...] = (0, 0)
    _default: tuple[float, ...] = (0, 0)
    _line_min: Optional[float] = None
    _line_max: Optional[float] = None
    _slider_min: float = 0
    _slider_max: float = 1
    _decimals: int = 4

    def _init_ui(self) -> None:
        # lines
        self.lines = []
        for i in range(self.multi_count):
            line = FloatLineEdit()
            line.value_changed.connect(self._line_value_changed)
            self.layout().addWidget(line)
            self.lines.append(line)

        # slider
        self.slider = FloatSlider()
        self.slider.value_changed.connect(self._slider_value_changed)
        # prevent any size changes when slider shows
        line_height = self.lines[0].minimumSizeHint().height()
        self.slider.setMaximumHeight(line_height)
        self.layout().addWidget(self.slider)
        self.layout().setStretch(self.multi_count, 1)

        # keep ratio button
        self.keep_ratio_button = RatioButton()
        self.keep_ratio_button.setMaximumSize(line_height, line_height)
        self.keep_ratio_button.toggled.connect(self.set_keep_ratio)
        self.layout().addWidget(self.keep_ratio_button)

        self.setFocusProxy(self.lines[0])

    def decimals(self) -> int:
        return self._decimals

    def set_decimals(self, decimals: int) -> None:
        self._decimals = decimals
        for line in self.lines:
            line.set_decimals(decimals)

    def line_min(self) -> float:
        return super().line_min()

    def line_max(self) -> float:
        return super().line_max()

    def slider_min(self) -> float:
        return super().slider_min()

    def slider_max(self) -> float:
        return super().slider_max()

    def set_line_min(self, line_min: float) -> None:
        super().set_line_min(line_min)

    def set_line_max(self, line_max: float) -> None:
        super().set_line_max(line_max)

    def set_slider_min(self, slider_min: float) -> None:
        super().set_slider_min(slider_min)

    def set_slider_max(self, slider_max: float) -> None:
        super().set_slider_max(slider_max)

    def set_value(self, value: tuple[float, ...]) -> None:
        super().set_value(value)

    def value(self) -> tuple[float, ...]:
        return super().value()


class PointParameter(MultiIntParameter):
    value_changed: QtCore.Signal = QtCore.Signal(QtCore.QPoint)

    _value: QtCore.QPoint = QtCore.QPoint(0, 0)
    _default: QtCore.QPoint = QtCore.QPoint(0, 0)
    _slider_visible: bool = False
    _ratio_visible: bool = False

    def set_value(self, value: QtCore.QPoint | list | tuple) -> None:
        super().set_value(value)

    def value(self) -> QtCore.QPoint:
        return super().value()

    def _cast_to_type(self, values: tuple[int, ...]) -> QtCore.QPoint:
        return QtCore.QPoint(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QPoint) -> tuple[int, ...]:
        return value.x(), value.y()


class PointFParameter(MultiFloatParameter):
    value_changed: QtCore.Signal = QtCore.Signal(QtCore.QPointF)

    _value: QtCore.QPointF = QtCore.QPointF(0, 0)
    _default: QtCore.QPointF = QtCore.QPointF(0, 0)
    _slider_visible: bool = False
    _ratio_visible: bool = False

    def set_value(self, value: QtCore.QPointF | list | tuple) -> None:
        super().set_value(value)

    def value(self) -> QtCore.QPointF:
        return super().value()

    def _cast_to_type(self, values: tuple[float, ...]) -> QtCore.QPointF:
        return QtCore.QPointF(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QPointF) -> tuple[float, ...]:
        return value.x(), value.y()


class SizeParameter(MultiIntParameter):
    value_changed: QtCore.Signal = QtCore.Signal(QtCore.QSize)

    _value: QtCore.QSize = QtCore.QSize(0, 0)
    _default: QtCore.QSize = QtCore.QSize(0, 0)

    def set_value(self, value: QtCore.QSize | list | tuple) -> None:
        super().set_value(value)

    def value(self) -> QtCore.QSize:
        return super().value()

    def _cast_to_type(self, values: tuple[int, ...]) -> QtCore.QSize:
        return QtCore.QSize(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QSize) -> tuple[int, ...]:
        return value.width(), value.height()


class SizeFParameter(MultiFloatParameter):
    value_changed: QtCore.Signal = QtCore.Signal(QtCore.QSizeF)

    _value: QtCore.QSizeF = QtCore.QSizeF(0, 0)
    _default: QtCore.QSizeF = QtCore.QSizeF(0, 0)

    def set_value(self, value: QtCore.QSizeF | list | tuple) -> None:
        super().set_value(value)

    def value(self) -> QtCore.QSizeF:
        return super().value()

    def _cast_to_type(self, values: tuple[float, ...]) -> QtCore.QSizeF:
        return QtCore.QSizeF(*values[:2])

    def _cast_to_tuple(self, value: QtCore.QSizeF) -> tuple[float, ...]:
        return value.width(), value.height()


class ColorParameter(MultiFloatParameter):
    multi_count = 3
    value_changed: QtCore.Signal = QtCore.Signal(QtGui.QColor)

    _value: QtGui.QColor = QtGui.QColor(0, 0, 0)
    _default: QtGui.QColor = QtGui.QColor(0, 0, 0)
    _color_min: Optional[float] = 0
    _color_max: Optional[float] = 1
    _decimals: int = 2

    def _init_ui(self) -> None:
        super()._init_ui()
        self.button = QtWidgets.QPushButton()
        self.button.clicked.connect(self.select_color)
        self.button.setFocusPolicy(QtCore.Qt.NoFocus)
        size = self.button.sizeHint()
        self.button.setMaximumWidth(size.height())
        self.layout().insertWidget(self.layout().count() - 1, self.button)

    def color_min(self) -> float:
        return self._color_min

    def color_max(self) -> float:
        return self._color_max

    def select_color(self) -> None:
        color = QtWidgets.QColorDialog.getColor(
            initial=self._value, options=QtWidgets.QColorDialog.DontUseNativeDialog
        )
        if color.isValid():
            super().set_value(color)
            self._set_line_values(self._cast_to_tuple(color))
            self._set_button_value(color)

    def set_color_min(self, color_min: float) -> None:
        self._color_min = color_min
        for line in self.lines:
            line.set_minimum(self._color_min)

    def set_color_max(self, color_max: float) -> None:
        self._color_max = color_max
        for line in self.lines:
            line.set_maximum(self._color_max)

    def set_value(self, value: QtGui.QColor | list | tuple) -> None:
        super().set_value(value)
        self._set_button_value(self._value)

    def value(self) -> QtGui.QColor:
        return super().value()

    def _cast_to_type(self, values: tuple[float, ...]) -> QtGui.QColor:
        return QtGui.QColor.fromRgbF(*values[:3])

    def _cast_to_tuple(self, value: QtGui.QColor) -> tuple[float, ...]:
        return value.getRgbF()[:3]

    def _line_value_changed(self, value: float) -> None:
        super()._line_value_changed(value)
        self._set_button_value(self._value)

    def _set_button_value(self, value: QtGui.QColor) -> None:
        self.button.setPalette(QtGui.QPalette(value))


class IntLineEdit(QtWidgets.QLineEdit):
    value_changed = QtCore.Signal(int)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._init_validator()

        self._abs_minimum = self.validator().bottom()
        self._abs_maximum = self.validator().top()
        self._minimum = self._abs_minimum
        self._maximum = self._abs_maximum
        self._value = 0

        self.commit_on_edit = False

        self.editingFinished.connect(self.commit)
        self.textEdited.connect(self._text_edit)

    def _init_validator(self) -> None:
        validator = IntValidator()
        self.setValidator(validator)

    def commit(self, update_text: bool = True) -> None:
        # strip padding
        value = self._text_to_value(self.text())
        if int(value) == value:
            value = int(value)
        if value != self._value:
            self._value = value
            self.value_changed.emit(value)
        if update_text:
            self.setText(str(value))

    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        if event.key() == QtCore.Qt.Key_Up:
            self._step(add=True)
            event.accept()
        elif event.key() == QtCore.Qt.Key_Down:
            self._step(add=False)
            event.accept()
        else:
            return super().keyPressEvent(event)

    def minimumSizeHint(self) -> QtCore.QSize:
        size = super().minimumSizeHint()
        size.setWidth(24)
        return size

    def sizeHint(self) -> QtCore.QSize:
        size = super().sizeHint()
        size.setWidth(60)
        return size

    def maximum(self) -> int:
        return self._maximum

    def minimum(self) -> int:
        return self._minimum

    def set_minimum(self, minimum: Optional[int]) -> None:
        if minimum is None:
            minimum = self._abs_minimum
        self._minimum = minimum
        self.validator().setBottom(minimum)

    def set_maximum(self, maximum: Optional[int]) -> None:
        if maximum is None:
            maximum = self._abs_maximum
        self._maximum = maximum
        self.validator().setTop(maximum)

    def set_value(self, value: int) -> None:
        text = self.validator().fixup(str(value))
        state, text_, pos_ = self.validator().validate(text, 0)
        if state == QtGui.QValidator.State.Acceptable:
            self.setText(text)
            self.commit()

    def value(self) -> int:
        return self._value

    def wheelEvent(self, event) -> None:
        delta = event.angleDelta()
        if delta.y() > 0:
            self._step(add=True)
        elif delta.y() < 0:
            self._step(add=False)
        event.accept()

    def _match_value_to_text(self, value: int, text: str, exponent: int) -> str:
        # exponent is for subclasses
        padding = len([t for t in text if t.isdigit()])
        if value < 0:
            padding += 1
        text = f'{value:0{padding}}'
        return text

    def _step(self, add: int) -> bool:
        self.setFocus()
        text = self.text() or '0'
        position = self.cursorPosition()
        if self.hasSelectedText():
            position = self.selectionStart()

        # check if cursor is on special character
        if position < len(text) and not text[position].isdigit():
            return False

        step_index = self._step_index(text, position)
        exponent = self._step_exponent(step_index)

        # perform step
        amount = 1 if add else -1
        step = amount * pow(10, exponent)
        value = self._value + step

        # preserve padding
        text = self._match_value_to_text(value, text, exponent)

        # validate before setting new text
        state, text_, pos_ = self.validator().validate(text, 0)
        if state != QtGui.QValidator.State.Acceptable:
            return False
        self.setText(text)

        # don't commit change to preserve padding
        self._value = value
        self.value_changed.emit(value)

        # get new position and set selection
        position = self._step_index_to_position(step_index, text)
        self.setSelection(position, 1)
        return True

    def _step_exponent(self, step_index: int) -> int:
        # convert cursor position to exponent
        exponent = step_index - 1
        return exponent

    def _step_index(self, text: str, position: int) -> int:
        # get step index relative to decimal point
        # this preserves position when number gets larger or changes plus/minus sign
        step_index = len(text) - position
        # if cursor is at end, edit first digit
        step_index = max(1, step_index)
        return step_index

    def _step_index_to_position(self, step_index: int, text: str) -> int:
        position = len(text) - step_index
        return position

    def _text_edit(self):
        if self.commit_on_edit:
            self.commit(update_text=False)

    # noinspection PyMethodMayBeStatic
    def _text_to_value(self, text: str) -> int:
        try:
            return int(text)
        except ValueError:
            return 0


class FloatLineEdit(IntLineEdit):
    value_changed = QtCore.Signal(float)

    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._decimals = self.validator().decimals()

    def _init_validator(self) -> None:
        validator = DoubleValidator()
        validator.setNotation(QtGui.QDoubleValidator.StandardNotation)
        self.setValidator(validator)

    def commit(self, update_text: bool = True) -> None:
        # strip padding
        value = self._text_to_value(self.text())
        if value != self._value:
            self._value = value
            self.value_changed.emit(value)
        if update_text:
            if int(value) == value:
                value = int(value)
            self.setText(str(value))

    def decimals(self) -> int:
        return self._decimals

    def set_decimals(self, value: int) -> None:
        self._decimals = value
        self.validator().setDecimals(value)

    def set_value(self, value: float) -> None:
        super().set_value(value)

    def value(self) -> float:
        return super().value()

    def _match_value_to_text(self, value: int, text: str, exponent: int) -> str:
        decimal_index = text.find('.')

        # preserve padding
        if decimal_index == -1:
            padding_decimal = 0
        else:
            padding_decimal = len(text) - 1 - decimal_index
            text = text[:decimal_index]

        # preserve padding if we switch to something like 1.001 > 1.000
        padding_decimal = max(padding_decimal, -exponent)
        padding_int = len([t for t in text if t.isdigit()])
        # account for minus sign
        if value < 0:
            padding_int += 1

        # padding_int needs to contain both padding for int and decimals
        padding_int += padding_decimal + 1 * bool(padding_decimal)

        value = round(value, padding_decimal)
        text = f'{value:0{padding_int}.{padding_decimal}f}'

        return text

    def _step_exponent(self, step_index: int) -> int:
        # convert cursor position to exponent
        exponent = step_index
        # if cursor is on the decimal then edit the first decimal
        if step_index >= 0:
            exponent = step_index - 1

        return exponent

    def _step_index(self, text: str, position: int) -> int:
        # get step index relative to decimal point
        # this preserves position when number gets larger or changes plus/minus sign
        decimal_index = text.find('.')
        if decimal_index == -1:
            step_index = len(text) - position
        else:
            step_index = decimal_index - position
        return step_index

    def _step_index_to_position(self, step_index: int, text: str) -> int:
        decimal_index = text.find('.')
        position = len(text) - step_index
        if decimal_index > -1:
            # if position is on decimal point, move to first decimal
            if step_index == 0:
                step_index = -1
            position = decimal_index - step_index
        return position

    # noinspection PyMethodMayBeStatic
    def _text_to_value(self, text: str) -> float:
        try:
            return float(text)
        except ValueError:
            return float(0)


class IntValidator(QtGui.QIntValidator):
    def fixup(self, text: str) -> str:
        text = str(super().fixup(text))
        text = text.replace(',', '')
        try:
            text = str(max(min(int(text), self.top()), self.bottom()))
        except ValueError:
            pass
        return text


class DoubleValidator(QtGui.QDoubleValidator):
    def fixup(self, text: str) -> str:
        try:
            float(text)
        except ValueError:
            characters = '+-01234567890.'
            text = [t for t in text if t in characters]

        try:
            value = float(text)
            value = min(max(value, self.bottom()), self.top())
            value = round(value, self.decimals())
            text = '{value:.{decimals}f}'.format(value=value, decimals=self.decimals())
            return text
        except (ValueError, TypeError):
            return text


class IntSlider(QtWidgets.QSlider):
    def __init__(
        self,
        orientation: QtCore.Qt.Orientation = QtCore.Qt.Horizontal,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(orientation, parent)
        self.setTickPosition(QtWidgets.QSlider.TicksBothSides)
        self.setFocusPolicy(QtCore.Qt.NoFocus)

        self.value_changed = self.valueChanged

    def set_minimum(self, value: int) -> None:
        self.setMinimum(value)
        self._update_steps()

    def set_maximum(self, value: int) -> None:
        self.setMaximum(value)
        self._update_steps()

    def set_value(self, value: int) -> None:
        self.setSliderPosition(value)

    def _exponent(self) -> int:
        # automatically adjust step size and tick interval based on slider range
        num_range = abs(self.maximum() - self.minimum())
        if num_range == 0:
            num_range = 1
        exponent = math.log10(num_range)

        # round exponent up or down with weighting towards down
        if exponent % 1 > 0.8:
            exponent = math.ceil(exponent)
        else:
            exponent = math.floor(exponent)
        return exponent

    def _update_steps(self) -> None:
        step = pow(10, max(self._exponent() - 2, 0))

        self.setSingleStep(step)
        self.setPageStep(step * 10)
        self.setTickInterval(step * 10)


class FloatSlider(IntSlider):
    _fvalue_changed: QtCore.Signal = QtCore.Signal(float)

    def __init__(
        self,
        orientation: QtCore.Qt.Orientation = QtCore.Qt.Horizontal,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(orientation, parent)
        super().valueChanged.connect(self._value_changed)
        self.value_changed = self._fvalue_changed

        self._minimum = super().minimum()
        self._maximum = super().maximum()

        self.setSingleStep(1)
        self.setPageStep(10)
        self.setTickInterval(10)

    def minimum(self) -> float:
        return self._minimum

    def maximum(self) -> float:
        return self._maximum

    def set_minimum(self, minimum: float) -> None:
        value = self.value()
        self._minimum = minimum
        self._update_steps()
        self.set_value(value)

    def set_maximum(self, maximum: float) -> None:
        value = self.value()
        self._maximum = maximum
        self._update_steps()
        self.set_value(value)

    def set_value(self, value: float) -> None:
        if math.isnan(value):
            return
        int_value = self._int(value)
        self.setSliderPosition(int_value)

    def value(self) -> float:
        value = super().value()
        float_value = self._float(value)
        return float_value

    def _int(self, value: float) -> int:
        # convert from float to int in slider scale
        try:
            percentage = (value - self._minimum) / (self._maximum - self._minimum)
        except ZeroDivisionError:
            return 0
        slider_range = super().maximum() - super().minimum()
        clamped_value = min(max(percentage, 0), 1) * slider_range + super().minimum()
        return int(clamped_value)

    def _float(self, value: int) -> float:
        # convert from int slider scale to float
        slider_range = super().maximum() - super().minimum()
        try:
            percentage = (value - super().minimum()) / slider_range
        except ZeroDivisionError:
            return float('nan')
        float_value = self._minimum + (self._maximum - self._minimum) * percentage
        return float_value

    def _update_steps(self) -> None:
        # find a value that brings the float range into an int range
        # with step size locked to 1 and 10
        step = pow(10, -(self._exponent() - 2))

        self.blockSignals(True)
        self.setMinimum(int(self._minimum * step))
        self.setMaximum(int(self._maximum * step))
        self.blockSignals(False)

    def _value_changed(self, value: int) -> None:
        value = self._float(value)
        if not math.isnan(value):
            self.value_changed.emit(value)


class RatioButton(BaseButton):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.setIcon(MaterialIcon('link_off', size=self._icon_size))
        self.setIcon(MaterialIcon('link', size=self._icon_size), True)
        self.setMaximumSize(QtCore.QSize(self._icon_size, self._icon_size))
        self.setCheckable(True)


class TextEdit(QtWidgets.QPlainTextEdit):
    editing_finished: QtCore.Signal = QtCore.Signal()

    def focusOutEvent(self, event: QtGui.QFocusEvent) -> None:
        self.editing_finished.emit()
        return super().focusOutEvent(event)

    def sizeHint(self) -> QtCore.QSize:
        size_hint = super().sizeHint()
        size_hint.setHeight(self.minimumSizeHint().height())
        return size_hint


__all__ = [
    'ParameterWidget',
    'IntParameter',
    'FloatParameter',
    'MultiIntParameter',
    'MultiFloatParameter',
    'PointParameter',
    'PointFParameter',
    'SizeParameter',
    'SizeFParameter',
    'StringParameter',
    'PathParameter',
    'BoolParameter',
    'EnumParameter',
    'ColorParameter',
]



from PySide2 import QtCore, QtGui, QtWidgets


class ElideLabel(QtWidgets.QLabel):
    def __init__(self, text='', parent=None):
        super().__init__(text, parent)

        self._text = text

    def minimumSizeHint(self) -> QtCore.QSize:
        size = super().minimumSizeHint()
        size.setWidth(size.height())
        return size

    def setText(self, text):
        super().setText(text)
        self._text = self.text()

    def _update_text(self):
        font_metrics = self.fontMetrics()
        width = self.contentsRect().width()
        text = font_metrics.elidedText(self._text, QtCore.Qt.ElideRight, width)
        super().setText(text)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self._update_text()



from PySide2 import QtWidgets, QtGui, QtCore


class ResizeGrip(QtWidgets.QWidget):
    def __init__(self, parent: QtWidgets.QWidget) -> None:
        super().__init__(parent)

        self.can_resize_vertical = True
        self.can_resize_horizontal = False

        self._resizing = False
        self._min_size = None
        self._start_position = QtCore.QPoint()
        self._start_size = QtCore.QSize()
        self._start_size_policy = None
        self._start_max_size = None

        self.setCursor(QtCore.Qt.SizeFDiagCursor)
        self.reset()
        parent.installEventFilter(self)

    @property
    def min_size(self) -> QtCore.QSize:
        if self._min_size is None:
            min_size = self.parent().minimumSize()
            min_size_hint = self.parent().minimumSizeHint()
            min_width = max(min_size.width(), min_size_hint.width(), self.width())
            min_height = max(min_size.height(), min_size_hint.height(), self.height())
            self._min_size = QtCore.QSize(min_width, min_height)
        return self._min_size

    def changeEvent(self, event: QtCore.QEvent) -> None:
        if event.type() == QtCore.QEvent.ParentChange and self.parent():
            self.reset()

    def eventFilter(self, obj, event: QtCore.QEvent) -> bool:
        if event.type() == QtCore.QEvent.Resize and obj == self.parent():
            self.reposition()
            self.resize_scroll_bars()
            return False
        return super().eventFilter(obj, event)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        painter = QtGui.QPainter(self)
        opt = QtWidgets.QStyleOptionSizeGrip()
        opt.initFrom(self)
        opt.corner = QtCore.Qt.BottomRightCorner
        self.style().drawControl(QtWidgets.QStyle.CE_SizeGrip, opt, painter, self)

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseDoubleClickEvent(event)
        if event.button() == QtCore.Qt.LeftButton:
            self.reset()

    def mousePressEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mousePressEvent(event)
        self._resizing = True
        self._start_size = self.parent().geometry().size()
        self._start_position = event.globalPos()
        if self._start_size_policy is None:
            self._start_size_policy = self.parent().sizePolicy()
        if self._start_max_size is None:
            self._start_max_size = self.parent().maximumSize()

    def mouseMoveEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseMoveEvent(event)
        if self._resizing:
            delta = event.globalPos() - self._start_position
            if self.can_resize_horizontal:
                width = self._start_size.width() + delta.x()
                width = max(width, self.min_size.width())
                self.parent().setFixedWidth(width)

            if self.can_resize_vertical:
                height = self._start_size.height() + delta.y()
                height = max(height, self.min_size.height())
                self.parent().setFixedHeight(height)

    def mouseReleaseEvent(self, event: QtGui.QMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        self._resizing = False

    def reset(self) -> None:
        # reset size
        size = self.parent().style().pixelMetric(QtWidgets.QStyle.PM_SizeGripSize)
        self.setFixedSize(size, size)

        # reset sizePolicy
        if self._start_size_policy is not None:
            policy = self._start_size_policy
            self._start_size_policy = None
            self.parent().setSizePolicy(policy)

        # reset maximumSize
        if self._start_max_size is not None:
            max_size = self._start_max_size
            self._start_max_size = None
            self.parent().setMaximumSize(max_size)

        # reset minimumSize
        self.parent().setMinimumSize(self.parent().minimumSizeHint())
        self._min_size = None

    def reposition(self) -> None:
        geometry = self.geometry()
        geometry.moveBottomRight(self.parent().contentsRect().bottomRight())
        self.setGeometry(geometry)

    def resize_scroll_bars(self) -> None:
        if isinstance(self.parent(), QtWidgets.QAbstractScrollArea):
            size = self.parent().contentsRect().size() - self.size()
            self.parent().horizontalScrollBar().setMaximumWidth(size.width())
            self.parent().verticalScrollBar().setMaximumHeight(size.height())



# initially https://github.com/marella/material-design-icons/ was added as a submodule
# but zip files of the repo on GitHub does not include submodules.
# to go around this issue icons are compiled to qt resource file using compile_icons.py
from __future__ import annotations

import enum

from PySide2 import QtWidgets, QtGui, QtCore
from PySide2.QtGui import QIcon, QPalette

from qt_extensions import icons_resource


def fill_pixmap(pixmap: QtGui.QPixmap, color: QtGui.QColor) -> QtGui.QPixmap:
    pixmap = QtGui.QPixmap(pixmap)
    painter = QtGui.QPainter(pixmap)
    painter.setCompositionMode(QtGui.QPainter.CompositionMode_SourceIn)
    painter.fillRect(pixmap.rect(), color)
    painter.end()
    return pixmap


class MaterialIcon(QIcon):
    class Style(enum.Enum):
        FILLED = 'filled'
        OUTLINED = 'outlined'
        ROUND = 'round'
        SHARP = 'sharp'
        TWO_TONE = 'two-tone'

    def __init__(
        self, name: str, style: Style | None = None, size: int | None = None
    ) -> None:
        super().__init__()

        # set pixmap
        if style is None:
            style = MaterialIcon.Style.OUTLINED

        self.name = name
        self._path = f':/material-design-icons/svg/{style.value}/{name}.svg'
        self._pixmap = QtGui.QPixmap(self._path)

        if size:
            self._pixmap = self._pixmap.scaled(QtCore.QSize(size, size))

        self._init_colors()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.name})'

    def _init_colors(self) -> None:
        # get palette colors
        app = QtWidgets.QApplication.instance()
        palette = app.palette()
        self._colors = {
            QIcon.On: {
                QIcon.Normal: None,
                QIcon.Active: None,
                QIcon.Disabled: None,
                QIcon.Selected: None,
            },
            QIcon.Off: {
                QIcon.Normal: palette.color(QPalette.Normal, QPalette.WindowText),
                QIcon.Active: None,
                QIcon.Disabled: palette.color(QPalette.Disabled, QPalette.WindowText),
                QIcon.Selected: None,
            },
        }

        for state, modes in self._colors.items():
            for mode, color in modes.items():
                if isinstance(color, QtGui.QColor):
                    self.set_color(color, mode, state)

    def add_icon(
        self,
        icon: QIcon,
        mode: QtGui.QIcon.Mode = QtGui.QIcon.Normal,
        state: QtGui.QIcon.State = QtGui.QIcon.Off,
    ):
        if isinstance(icon, MaterialIcon):
            pixmap = icon._pixmap
        else:
            pixmap = icon.pixmap(self._pixmap.size())
        self.addPixmap(pixmap, mode, state)

    def pixmap(
        self,
        extent: int = 0,
        mode: QtGui.QIcon.Mode = QtGui.QIcon.Normal,
        state: QtGui.QIcon.State = QtGui.QIcon.Off,
        color: QtGui.QColor | None = None,
    ) -> QtGui.QPixmap:
        if extent:
            # pixmap = QtGui.QPixmap(self._path).scaledToWidth(extent)
            pixmap = QtGui.QIcon(self._path).pixmap(extent)
        else:
            pixmap = self._pixmap

        if color is None:
            color = self._colors[state][mode]
            if color is None:
                color = self._colors[QIcon.Off][QIcon.Normal]

        return fill_pixmap(pixmap, color)

    def set_color(
        self,
        color: QtGui.QColor,
        mode: QtGui.QIcon.Mode = QIcon.Normal,
        state: QtGui.QIcon.State = QIcon.Off,
    ):
        pixmap = fill_pixmap(self._pixmap, color)
        self.addPixmap(pixmap, mode, state)



from __future__ import annotations

import os
import re
from collections.abc import Iterable


def title(text: str) -> str:
    text = re.sub(r'([a-z0-9])([A-Z])', r'\g<1> \g<2>', text).replace('_', ' ').title()
    return text


def unique_name(name: str, existing_names: Iterable[str]) -> str:
    for existing_name in existing_names:
        if name.casefold() == existing_name.casefold():
            match = re.search(r'(.*?)(\d+)$', name)
            if match:
                number = int(match.group(2))
                name = match.group(1)
            else:
                number = 0
            name = f'{name}{number + 1}'
            return unique_name(name, existing_names)
    return name


def unique_path(path: str) -> str:
    parent_path = os.path.dirname(path)
    try:
        items = os.listdir(parent_path)
    except OSError:
        return path

    basename = os.path.basename(path)
    name, ext = os.path.splitext(basename)
    names = []
    for item in items:
        item_name, item_ext = os.path.splitext(item)
        if ext == item_ext:
            names.append(item_name)
    basename = unique_name(name, names) + ext
    path = os.path.join(parent_path, basename)
    return path



from __future__ import annotations

import logging
import os
from functools import partial

from PySide2 import QtWidgets, QtCore, QtGui

from qt_extensions import theme
from qt_extensions.button import CheckBoxButton
from qt_extensions.icons import MaterialIcon


SUCCESS = 25

logging.addLevelName(SUCCESS, 'SUCCESS')


class Sender(QtCore.QObject):
    signal = QtCore.Signal(object)


class Handler(logging.Handler):
    def __init__(self, level: int = logging.NOTSET) -> None:
        super().__init__(level)

        self._sender = Sender()
        self.record_logged = self._sender.signal

    def emit(self, record: logging.LogRecord) -> None:
        self.record_logged.emit(record)


class LogCache(QtCore.QObject):
    added: QtCore.Signal = QtCore.Signal(logging.LogRecord)
    cleared: QtCore.Signal = QtCore.Signal()

    def __init__(self, parent: QtCore.QObject | None = None) -> None:
        super().__init__(parent)

        self.records = []
        self.handler = Handler()
        self.handler.record_logged.connect(self.add)

    def add(self, record: logging.LogRecord) -> None:
        self.records.append(record)
        self.added.emit(record)

    def clear(self) -> None:
        self.records = []
        self.cleared.emit()

    def connect_logger(self, logger: logging.Logger) -> None:
        logger.setLevel(logging.NOTSET)
        logger.addHandler(self.handler)

    def save(self, filename: str) -> None:
        formatter = logging.Formatter(
            fmt='[{asctime}][{levelname: <8}][{name}] {message}',
            datefmt='%I:%M:%S%p',
            style='{',
        )

        with open(filename, 'w') as f:
            text = (f'{formatter.format(record)}\n' for record in self.records)
            f.writelines(text)


class LogViewer(QtWidgets.QWidget):
    def __init__(
        self, cache: LogCache | None = None, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self._cache = None
        self._cache_connected = False
        self._error_color = theme.Color('error').name()
        self._warning_color = theme.Color('warning').name()
        self._success_color = theme.Color('success').name()
        self._error_count = 0
        self._warning_count = 0
        self._last_save_path = os.path.expanduser('~')
        self._names = set()
        self._levels = set()

        fmt = (
            '[{asctime}]<font color="{color}"><b>[{levelname: <8}]</b></font> {message}'
        )
        self.formatter = logging.Formatter(
            fmt=fmt,
            datefmt='%I:%M:%S%p',
            style='{',
        )

        self._init_ui()

        self.set_levels((logging.ERROR, logging.WARNING))

        if cache:
            self.set_cache(cache)

    def _init_ui(self) -> None:
        self.setWindowTitle('Log Viewer')

        self.setLayout(QtWidgets.QVBoxLayout())

        self.text_edit = QtWidgets.QPlainTextEdit()
        self.text_edit.setReadOnly(True)
        self.text_edit.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)

        font = QtGui.QFont('Roboto Mono')
        font.setStyleHint(QtGui.QFont.Monospace)
        self.text_edit.setFont(font)

        self.layout().addWidget(self.text_edit)

        # toolbar
        self.toolbar = QtWidgets.QToolBar()
        size = self.style().pixelMetric(QtWidgets.QStyle.PM_SmallIconSize)
        self.toolbar.setIconSize(QtCore.QSize(size, size))
        self.layout().insertWidget(0, self.toolbar)

        # level buttons
        button = CheckBoxButton('Error', color='error')
        button.setFlat(True)
        button.setFocusPolicy(QtCore.Qt.NoFocus)
        button.toggled.connect(partial(self._level_toggle, logging.ERROR))
        self.toolbar.addWidget(button)
        self._error_button = button

        button = CheckBoxButton('Warning', color='warning')
        button.setFlat(True)
        button.setFocusPolicy(QtCore.Qt.NoFocus)
        button.toggled.connect(partial(self._level_toggle, logging.WARNING))
        self.toolbar.addWidget(button)
        self._warning_button = button

        button = CheckBoxButton('Info')
        button.setFlat(True)
        button.setFocusPolicy(QtCore.Qt.NoFocus)
        button.toggled.connect(partial(self._level_toggle, logging.INFO))
        self.toolbar.addWidget(button)
        self._info_button = button

        button = CheckBoxButton('Debug')
        button.setFlat(True)
        button.setFocusPolicy(QtCore.Qt.NoFocus)
        button.toggled.connect(partial(self._level_toggle, logging.DEBUG))
        self.toolbar.addWidget(button)
        self._debug_button = button

        # stretch
        spacer = QtWidgets.QWidget()
        spacer.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        self.toolbar.addWidget(spacer)

        # actions
        action = QtWidgets.QAction(self)
        action.setText('Filter')
        icon = MaterialIcon('filter_alt')
        action.setIcon(icon)
        action.triggered.connect(self._show_filter_menu)
        self.toolbar.addAction(action)
        self._filter_action = action

        action = QtWidgets.QAction(self)
        action.setText('Wrap Lines')
        action.setCheckable(True)
        icon = MaterialIcon('wrap_text')
        action.setIcon(icon)
        action.toggled.connect(self._wrap_text)
        self.toolbar.addAction(action)
        self._wrap_action = action

        action = QtWidgets.QAction(self)
        action.setText('Save')
        icon = MaterialIcon('save')
        action.setIcon(icon)
        action.triggered.connect(self.save)
        self.toolbar.addAction(action)

        action = QtWidgets.QAction(self)
        action.setText('Clear')
        icon = MaterialIcon('backspace')
        action.setIcon(icon)
        self.toolbar.addAction(action)
        self._clear_action = action

    @property
    def error_count(self) -> int:
        return self._error_count

    @error_count.setter
    def error_count(self, value: int) -> None:
        self._error_count = value
        self._error_button.setText(f'Error: {value}')

    @property
    def warning_count(self) -> int:
        return self._warning_count

    @warning_count.setter
    def warning_count(self, value: int) -> None:
        self._warning_count = value
        self._warning_button.setText(f'Warning: {value}')

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self._disconnect_cache()
        super().closeEvent(event)

    def showEvent(self, event: QtGui.QShowEvent) -> None:
        self._connect_cache()
        super().showEvent(event)
        self.refresh()

    def add_name(self, name: str, add: bool = True) -> None:
        if add:
            self._names.add(name)
        else:
            try:
                self._names.remove(name)
            except KeyError:
                pass
        self.refresh()

    def add_record(self, record: logging.LogRecord, count: bool = True) -> None:
        if self._names and not record.name.startswith(tuple(self._names)):
            return

        # NOTE: python 3.9 does not allow defaults in the formatter
        record.color = ''

        if record.levelno >= logging.ERROR:
            if count:
                self.error_count += 1
            if logging.ERROR not in self._levels:
                return
            record.color = self._error_color
        elif record.levelno >= logging.WARNING:
            if count:
                self.warning_count += 1
            if logging.WARNING not in self._levels:
                return
            record.color = self._warning_color
        elif record.levelno == SUCCESS:
            if logging.INFO not in self._levels:
                return
            record.color = self._success_color
        elif record.levelno >= logging.INFO:
            if logging.INFO not in self._levels:
                return
        elif record.levelno >= logging.DEBUG:
            if logging.DEBUG not in self._levels:
                return
        else:
            return

        message = self.formatter.format(record)
        if record.exc_info:
            lines = message.split('\n')
            try:
                header = lines.pop(0)
                trace = '\n'.join(lines)
                message = f'{header}\n<font color="{self._error_color}">{trace}</font>'
            except IndexError:
                pass
        inner_html = f'<pre><code data-lang="python">{message}</code></pre>'

        self.text_edit.appendHtml(inner_html)

    def cache(self) -> LogCache | None:
        return self._cache

    def clear(self) -> None:
        self.text_edit.clear()
        self.error_count = 0
        self.warning_count = 0

    def refresh(self) -> None:
        self.clear()
        if self._cache:
            for record in self._cache.records:
                self.add_record(record)

    def save(self) -> None:
        if not self._cache:
            return
        filename, selected_filter = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption='Save File',
            dir=self._last_save_path,
            filter='*.log',
        )

        if filename:
            path = os.path.dirname(filename)
            self._last_save_path = path
            if not os.path.exists(path):
                os.makedirs(path)
            self._cache.save(filename)

    def set_cache(self, cache: LogCache) -> None:
        self._disconnect_cache()
        self._cache = cache
        self._clear_action.triggered.connect(self._cache.clear)
        if self.isVisible():
            self._connect_cache()

    def set_levels(self, levels: set[int] | list[int] | tuple[int, ...]) -> None:
        self._levels = set(levels)
        self._error_button.setChecked(logging.ERROR in self._levels)
        self._warning_button.setChecked(logging.WARNING in self._levels)
        self._info_button.setChecked(logging.INFO in self._levels)
        self._debug_button.setChecked(logging.DEBUG in self._levels)

    def set_names(self, names: set[str] | list[str] | tuple[str]) -> None:
        self._names = set(names)
        self.refresh()

    def set_state(self, state: dict) -> None:
        values = {
            'names': set(),
            'levels': {logging.ERROR, logging.WARNING, logging.INFO},
            'wrap': False,
        }
        values.update(state)

        self.set_levels(values['levels'])
        self.set_names(values['names'])
        self._wrap_action.setChecked(values['wrap'])

    def state(self) -> dict:
        log_viewer_state = {
            'names': self._names,
            'levels': self._levels,
            'wrap': self._wrap_action.isChecked(),
        }
        return log_viewer_state

    def _connect_cache(self) -> None:
        if self._cache and not self._cache_connected:
            self._cache.added.connect(self.add_record)
            self._cache.cleared.connect(self.clear)
            self._cache_connected = True

    def _disconnect_cache(self) -> None:
        if self._cache:
            try:
                self._cache.added.disconnect(self.add_record)
                self._cache.cleared.disconnect(self.clear)
            except RuntimeError:
                pass
            self._cache_connected = False
            self.clear()

    def _filter(self) -> None:
        self.text_edit.clear()
        if self._cache:
            for record in self._cache.records:
                self.add_record(record, count=False)

    def _level_toggle(self, level: int, checked: bool) -> None:
        if checked:
            self._levels.add(level)
        else:
            try:
                self._levels.remove(level)
            except KeyError:
                pass
        self._filter()

    def _show_filter_menu(self) -> None:
        if not self._cache:
            return
        widget = self.toolbar.widgetForAction(self._filter_action)
        relative_pos = widget.rect().topRight()
        relative_pos.setX(relative_pos.x() + 2)
        position = widget.mapToGlobal(relative_pos)

        names = set()
        for record in self._cache.records:
            name = record.name.split('.')[0]
            names.add(name)

        for name in self._names:
            names.add(name)

        menu = QtWidgets.QMenu(self)
        for name in names:
            action = QtWidgets.QAction(self)
            action.setText(name)
            action.setCheckable(True)
            action.toggled.connect(partial(self.add_name, name))
            menu.addAction(action)

            if name in self._names:
                action.setChecked(True)
        menu.exec_(position)

    def _wrap_text(self, checked: bool) -> None:
        if checked:
            self.text_edit.setLineWrapMode(QtWidgets.QPlainTextEdit.WidgetWidth)
        else:
            self.text_edit.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
            self.text_edit.horizontalScrollBar().setValue(0)


class LogBar(QtWidgets.QWidget):
    def __init__(
        self, cache: LogCache | None = None, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self._cache = None
        self._viewer = None

        self.current_message = logging.makeLogRecord({'levelno': logging.NOTSET})
        self.formatter = logging.Formatter(fmt='[{levelname}] {message}', style='{')
        self.level = logging.WARNING
        self.names = set()

        self._init_ui()

        if cache:
            self.set_cache(cache)

    def _init_ui(self) -> None:
        # colors
        self._critical_color = theme.Color('critical')
        self._error_color = theme.Color('error')
        self._warning_color = theme.Color('warning')
        self._success_color = theme.Color('success')

        # log icons
        self._critical_icon = MaterialIcon('report')
        self._critical_icon.set_color(self._critical_color)
        self._error_icon = MaterialIcon('error')
        self._error_icon.set_color(self._error_color)
        self._warning_icon = MaterialIcon('warning')
        self._warning_icon.set_color(self._warning_color)
        self._info_icon = MaterialIcon('article')
        self._success_icon = MaterialIcon('check_circle')
        self._success_icon.set_color(self._success_color)

        # layout
        size_policy = self.sizePolicy()
        size_policy.setVerticalPolicy(QtWidgets.QSizePolicy.Maximum)
        self.setSizePolicy(size_policy)

        layout = QtWidgets.QHBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(QtCore.QMargins())

        # message
        message_layout = QtWidgets.QHBoxLayout()
        message_layout.setContentsMargins(2, 2, 2, 2)
        message_layout.setSpacing(0)

        # message button
        self.log_button = QtWidgets.QToolButton(self)
        self.log_button.setIcon(self._info_icon)
        self.log_button.setAutoRaise(True)
        self.log_button.setFocusPolicy(QtCore.Qt.NoFocus)
        self.log_button.pressed.connect(self.open_viewer)

        message_layout.addWidget(self.log_button)

        # message line
        self.message_line = QtWidgets.QLineEdit(self)
        self.message_line.setReadOnly(True)
        self.message_line.setFocusPolicy(QtCore.Qt.NoFocus)

        font = QtGui.QFont('Roboto Mono')
        font.setStyleHint(QtGui.QFont.Monospace)
        self.message_line.setFont(font)
        message_layout.addWidget(self.message_line)

        layout.addLayout(message_layout)
        layout.setStretch(0, 1)

        # size grip
        self.size_grip = QtWidgets.QSizeGrip(self)
        layout.addWidget(
            self.size_grip, 0, QtCore.Qt.AlignBottom | QtCore.Qt.AlignRight
        )

    def add_widget(self, widget: QtWidgets.QWidget) -> None:
        count = self.layout().count()
        self.layout().insertWidget(count - 1, widget)

    def clear_message(self) -> None:
        self.show_message('', level=logging.NOTSET, force=True)

    def open_viewer(self) -> None:
        if self._viewer is None:
            self._viewer = LogViewer(self._cache, parent=self)
            self._viewer.setWindowFlag(QtCore.Qt.Dialog)
        self._viewer.show()

    def remove_widget(self, widget: QtWidgets.QWidget) -> None:
        self.layout().removeWidget(widget)

    def set_cache(self, cache: LogCache) -> None:
        self._cache = cache
        self._cache.added.connect(self.show_record)
        self._cache.cleared.connect(self.clear_message)

    def show_message(
        self, message: str, level: int = logging.INFO, force=False
    ) -> None:
        if not force:
            if level < self.current_message.levelno or level < self.level:
                return

        if level == SUCCESS:
            self.log_button.setIcon(self._success_icon)
            color = self._success_color
        elif level >= logging.CRITICAL:
            self.log_button.setIcon(self._critical_icon)
            color = self._critical_color
        elif level >= logging.ERROR:
            self.log_button.setIcon(self._error_icon)
            color = self._error_color
        elif level >= logging.WARNING:
            self.log_button.setIcon(self._warning_icon)
            color = self._warning_color
        else:
            self.log_button.setIcon(self._info_icon)
            color = self.palette().color(QtGui.QPalette.Window)

        self.message_line.setText(message)
        self.message_line.setCursorPosition(0)
        palette = self.message_line.palette()
        palette.setColor(QtGui.QPalette.Window, color)
        self.message_line.setPalette(palette)

        self.current_message = logging.makeLogRecord({'msg': message, 'levelno': level})

    def show_record(self, record: logging.LogRecord) -> None:
        if record is None or (
            self.names and not record.name.startswith(tuple(self.names))
        ):
            return
        message = self.formatter.format(record)
        message = message.split('\n')[-1]
        self.show_message(message, level=record.levelno)



from __future__ import annotations

import enum

from PySide2 import QtWidgets, QtCore, QtGui
from qt_extensions.icons import MaterialIcon


class CollapsibleBox(QtWidgets.QFrame):
    class Style(enum.Enum):
        NONE = 0
        SIMPLE = 1
        BUTTON = 2

    NONE = Style.NONE
    SIMPLE = Style.SIMPLE
    BUTTON = Style.BUTTON

    def __init__(self, title: str, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)

        self._maximum_height = self.maximumHeight()
        self._collapsed = False
        self._checkable = False
        self._collapsible = False
        self._style = CollapsibleBox.NONE

        self._title_label = None
        self._expand_label = None
        self._expand_more_icon = MaterialIcon('chevron_right')
        self._expand_less_icon = MaterialIcon('expand_more')
        self._menu_button = None

        self.frame = None
        self.header = None
        self._init_ui()

        self.set_title(title)

    def _init_ui(self) -> None:
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum
        )

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setSpacing(0)
        super().setLayout(self._layout)

        # header
        self.header = QtWidgets.QWidget()
        self.header.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.header.installEventFilter(self)
        self.header.setBackgroundRole(QtGui.QPalette.Base)
        self._layout.addWidget(self.header)

        header_layout = QtWidgets.QHBoxLayout()
        self.header.setLayout(header_layout)

        self._expand_label = QtWidgets.QLabel(self.header)
        self._expand_label.setVisible(False)
        header_layout.addWidget(self._expand_label)

        self.checkbox = QtWidgets.QCheckBox(self.header)
        self.checkbox.setVisible(False)
        header_layout.addWidget(self.checkbox)

        self._title_label = QtWidgets.QLabel(self.header)
        header_layout.addWidget(self._title_label)
        header_layout.addStretch()

        self._menu_button = QtWidgets.QToolButton(self.header)
        self._menu_button.setIcon(MaterialIcon('menu'))
        self._menu_button.setAutoRaise(True)
        self._menu_button.setVisible(False)
        self._menu_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self._menu_button.setFocusPolicy(QtCore.Qt.NoFocus)
        self._menu_button.pressed.connect(self._show_menu)
        size = self.header.style().pixelMetric(QtWidgets.QStyle.PM_SmallIconSize)
        self._menu_button.setMaximumSize(QtCore.QSize(size, size))
        header_layout.addWidget(self._menu_button)

        self._update_icon()

        # frame
        self.frame = QtWidgets.QFrame(self)
        self._layout.addWidget(self.frame)
        self._layout.setStretch(1, 1)

    def __repr__(self):
        return f'{self.__class__.__name__}({repr(self.title())})'

    def actionEvent(self, event: QtGui.QActionEvent) -> None:
        self._menu_button.setVisible(bool(self.actions()))
        super().actionEvent(event)

    def enterEvent(self, event: QtCore.QEvent) -> None:
        self.update()
        super().enterEvent(event)

    def eventFilter(self, watched: QtCore.QObject, event: QtCore.QEvent) -> bool:
        if watched == self.header and self._collapsible:
            if (
                event.type() == QtCore.QEvent.MouseButtonPress
                and event.button() == QtCore.Qt.LeftButton
            ):
                self.header.setAutoFillBackground(True)
            if (
                event.type() == QtCore.QEvent.MouseButtonRelease
                and event.button() == QtCore.Qt.LeftButton
            ):
                self.set_collapsed(not self._collapsed)
                self.header.setAutoFillBackground(False)
        return super().eventFilter(watched, event)

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        self.update()
        super().leaveEvent(event)

    def layout(self) -> QtWidgets.QLayout:
        return self.frame.layout()

    def setLayout(self, layout: QtWidgets.QLayout) -> None:
        self.frame.setLayout(layout)

    def paintEvent(self, event: QtGui.QPaintEvent) -> None:
        if not self._style == CollapsibleBox.BUTTON:
            return super().paintEvent(event)

        option = QtWidgets.QStyleOptionButton()
        option.initFrom(self)

        if not self._collapsed:
            option.state |= QtWidgets.QStyle.State_Sunken
        if self._collapsed and self.underMouse():
            option.state |= QtWidgets.QStyle.State_MouseOver

        painter = QtGui.QPainter(self)
        style = self.style()
        style.drawPrimitive(
            QtWidgets.QStyle.PE_PanelButtonCommand, option, painter, self
        )

    def setMaximumHeight(self, maxh: int) -> None:
        self._maximum_height = maxh
        super().setMaximumHeight(maxh)

    def setMaximumSize(self, size: QtCore.QSize) -> None:
        self._maximum_height = size.height()
        super().setMaximumSize(size)

    def box_style(self) -> Style:
        return self._style

    def checkable(self) -> bool:
        return self._checkable

    def checked(self) -> bool:
        return self._checkable and self.checkbox.isChecked()

    def collapsible(self) -> bool:
        return self._collapsible

    def collapsed(self) -> bool:
        return self._collapsible and self._collapsed

    def title(self) -> str:
        return self._title_label.text()

    def set_actions(self, actions: list[QtWidgets.QAction]) -> None:
        for action in self.actions():
            self.removeAction(action)
        self.addActions(actions)

    def set_checkable(self, checkable: bool) -> None:
        self._checkable = checkable
        self.checkbox.setVisible(checkable)

    def set_checked(self, checked: bool) -> None:
        if self.checkable():
            self.checkbox.setChecked(checked)

    def set_collapsible(self, collapsible: bool) -> None:
        if collapsible:
            margins = QtCore.QMargins(4, 4, 4, 4)
        else:
            self.set_collapsed(False)
            style = self.header.style()
            left = style.pixelMetric(QtWidgets.QStyle.PM_LayoutLeftMargin)
            top = style.pixelMetric(QtWidgets.QStyle.PM_LayoutTopMargin)
            right = style.pixelMetric(QtWidgets.QStyle.PM_LayoutRightMargin)
            bottom = style.pixelMetric(QtWidgets.QStyle.PM_LayoutBottomMargin)
            margins = QtCore.QMargins(left, top, right, bottom)

        self._collapsible = collapsible
        self.header.layout().setContentsMargins(margins)
        self._expand_label.setVisible(collapsible)

    def set_collapsed(self, collapsed: bool) -> None:
        if self.collapsible():
            self._collapsed = collapsed
            self.frame.setMaximumHeight(0 if collapsed else self._maximum_height)
            self._update_icon()

    def set_box_style(self, style: Style) -> None:
        self._style = style
        if style == CollapsibleBox.Style.SIMPLE:
            self.setFrameShape(QtWidgets.QFrame.StyledPanel)
        else:
            self.setFrameShape(QtWidgets.QFrame.NoFrame)
        if style == CollapsibleBox.Style.BUTTON:
            self._layout.setContentsMargins(2, 2, 2, 2)
        else:
            self._layout.setContentsMargins(0, 0, 0, 0)

    def set_title(self, title: str) -> None:
        self._title_label.setText(title)

    def _show_menu(self) -> None:
        relative_pos = self._menu_button.rect().topRight()
        relative_pos.setX(relative_pos.x() + 2)
        position = self._menu_button.mapToGlobal(relative_pos)

        menu = QtWidgets.QMenu(self)
        menu.addActions(self.actions())
        menu.exec_(position)

        self._menu_button.setDown(False)

    def _update_icon(self) -> None:
        icon = self._expand_more_icon if self._collapsed else self._expand_less_icon
        size = self.style().pixelMetric(QtWidgets.QStyle.PM_ButtonIconSize)
        self._expand_label.setPixmap(icon.pixmap(size))



from __future__ import annotations

import itertools
import typing
from collections.abc import Iterable
from functools import partial

from PySide2 import QtWidgets, QtCore, QtGui

from qt_extensions import helper
from qt_extensions.scrollarea import VerticalScrollArea
from qt_extensions.parameters import ParameterWidget
from qt_extensions.box import CollapsibleBox


class ParameterToggle(QtWidgets.QCheckBox):
    def __init__(self, name: str, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.name = name

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.name)})'

    def set_value(self, value: bool) -> None:
        self.setChecked(value)

    def value(self) -> bool:
        return self.isChecked()


class ParameterToolTip(QtWidgets.QFrame):
    def __init__(
        self, widget: ParameterWidget, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(parent)

        self.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.setAutoFillBackground(True)

        palette = self.palette()
        palette.setColor(QtGui.QPalette.Window, palette.color(QtGui.QPalette.Base))
        self.setPalette(palette)

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSizeConstraint(QtWidgets.QLayout.SetFixedSize)

        title = QtWidgets.QLabel(widget.label(), self)
        font = title.font()
        font.setBold(True)
        title.setFont(font)
        self.layout().addWidget(title)

        separator = QtWidgets.QFrame(self)
        separator.setFrameShape(QtWidgets.QFrame.HLine)
        self.layout().addWidget(separator)

        typ = type(widget).__name__.replace('Parameter', '')
        detail = QtWidgets.QLabel(f'Parameter: {widget.name()} ({typ})', self)
        self.layout().addWidget(detail)

        tooltip = QtWidgets.QLabel(widget.tooltip(), self)
        # tooltip.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)
        tooltip.setWordWrap(True)
        tooltip.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignLeft)
        self.layout().addWidget(tooltip)

    def focusOutEvent(self, event: QtCore.QEvent) -> None:
        self.hide()

    def leaveEvent(self, event: QtCore.QEvent) -> None:
        self.hide()


class ParameterLabel(QtWidgets.QLabel):
    def __init__(
        self, widget: ParameterWidget, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(widget.label(), parent)

        self._tooltip: ParameterToolTip | None = None
        self._widget = widget

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.text())})'

    def enterEvent(self, event: QtCore.QEvent) -> None:
        if self._widget.tooltip():
            QtCore.QTimer.singleShot(600, self.show_tooltip)
        super().enterEvent(event)

    def show_tooltip(self):
        global_position = QtGui.QCursor.pos()
        if self.geometry().contains(self.parent().mapFromGlobal(global_position)):
            if self._tooltip is None:
                self._tooltip = ParameterToolTip(self._widget, parent=self.window())
            self._tooltip.move(self.window().mapFromGlobal(global_position))
            self._tooltip.show()


class ParameterTabWidget(QtWidgets.QTabWidget):
    def __init__(self, parent: QtWidgets.QWidget | None = None) -> None:
        super().__init__(parent)
        self.tabs: dict[str, ParameterForm] = {}


class ParameterBox(CollapsibleBox):
    def __init__(
        self, title: str, form: ParameterForm, parent: QtWidgets.QWidget | None = None
    ) -> None:
        super().__init__(title, parent)
        self.form = form

        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().addWidget(form)


class ParameterForm(QtWidgets.QWidget):
    actions_changed: QtCore.Signal = QtCore.Signal(list)
    parameter_changed: QtCore.Signal = QtCore.Signal(ParameterWidget)

    # require unique names in the whole hierarchy
    unique_hierarchical_names: bool = False

    # return values as nested dict or flat hierarchy
    create_hierarchy: bool = True

    def __init__(
        self,
        name: str | None = None,
        root: ParameterForm | None = None,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        self._widgets: dict[str, ParameterWidget] = {}

        self.name = name
        self.root = root or self

        self.setLayout(QtWidgets.QGridLayout())

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.name)})'

    @property
    def grid_layout(self) -> QtWidgets.QGridLayout:
        layout = self.layout()
        if not isinstance(layout, QtWidgets.QGridLayout):
            raise RuntimeError('Layout needs to be QGridLayout')
        return layout

    def actionEvent(self, event: QtGui.QActionEvent) -> None:
        super().actionEvent(event)
        self.actions_changed.emit(self.actions())

    def add_group(self, name: str, label: str = None) -> ParameterBox:
        form = self._create_form(name)
        form.parameter_changed.connect(self.parameter_changed.emit)
        label = label or helper.title(name)
        box = ParameterBox(label, form, self)
        box.set_collapsible(True)
        box.set_collapsed(True)

        # update CollapsibleBox with actions for the menu
        form.actions_changed.connect(box.set_actions)

        self.add_widget(box)
        return box

    def add_layout(
        self,
        layout: QtWidgets.QLayout,
        column: int = 1,
        row_span: int = 1,
        column_span: int = -1,
    ) -> QtWidgets.QLayout:
        grid_layout = self.grid_layout
        row = grid_layout.rowCount() - 1
        grid_layout.addLayout(layout, row, column, row_span, column_span)
        self._update_stretch()
        return layout

    def add_parameter(
        self, widget: ParameterWidget, checkable: bool = False
    ) -> ParameterWidget:
        name = widget.name()
        self._validate_name(name)

        self._widgets[name] = widget

        layout = self.grid_layout
        row = layout.rowCount() - 1

        # label
        if widget.label():
            label = ParameterLabel(widget, self)
            column = 1
            layout.addWidget(label, row, column)
            widget.enabled_changed.connect(label.setEnabled)
            widget.enabled_changed.emit(widget.isEnabled())

        # widget
        column = 2
        layout.addWidget(widget, row, column)
        widget.value_changed.connect(lambda: self.parameter_changed.emit(widget))

        # checkbox
        if checkable:
            checkbox_name = f'{name}_enabled'
            checkbox = ParameterToggle(checkbox_name)
            column = 0
            layout.addWidget(checkbox, row, column)
            checkbox.toggled.connect(partial(self._set_widget_row_enabled, checkbox))
            checkbox.toggled.connect(lambda: self.parameter_changed.emit(checkbox))
            checkbox.set_value(False)
            self._set_widget_row_enabled(checkbox, False)

            self._widgets[checkbox_name] = checkbox

        self._update_stretch()
        return widget

    def add_separator(self) -> QtWidgets.QFrame:
        line = QtWidgets.QFrame(self)
        line.setFixedHeight(1)
        line.setFrameShape(QtWidgets.QFrame.StyledPanel)
        line.setFrameShadow(QtWidgets.QFrame.Sunken)

        self.add_widget(line)
        return line

    def add_tab_group(
        self, names: Iterable[str], labels: Iterable[str] = None
    ) -> ParameterTabWidget:
        tab_widget = ParameterTabWidget(self)
        tab_widget.tabs = {}
        labels = labels or []
        for name, label in itertools.zip_longest(names, labels):
            form = self._create_form(name)
            form.parameter_changed.connect(self.parameter_changed.emit)
            label = label or helper.title(name)
            tab_widget.addTab(form, label)
            tab_widget.tabs[name] = form

        self.add_widget(tab_widget)
        return tab_widget

    def add_widget(
        self, widget: QtWidgets.QWidget, column: int = 1, column_span: int = -1
    ) -> QtWidgets.QWidget:
        layout = self.grid_layout
        row = layout.rowCount() - 1
        row_span = 1
        layout.addWidget(widget, row, column, row_span, column_span)
        self._update_stretch()
        return widget

    def groups(self) -> dict:
        # create hierarchical dict of boxes, groups and tab widgets
        layout = self.grid_layout

        group_dict = {}
        for index in range(layout.count()):
            widget = layout.itemAt(index).widget()
            if isinstance(widget, ParameterBox):
                if widget.form:
                    group_dict[widget] = widget.form.groups()
            elif isinstance(widget, ParameterForm):
                group_dict.update(widget.groups())
            elif isinstance(widget, ParameterTabWidget):
                for form in widget.tabs.values():
                    group_dict[form] = form.groups()

        return group_dict

    def reset(self, widgets: dict[str, ParameterWidget] | None = None) -> None:
        if widgets is None:
            widgets = self.widgets()

        for widget in widgets.values():
            if isinstance(widget, ParameterWidget):
                widget.set_value(widget.default())
            elif isinstance(widget, dict):
                self.reset(widget)

    def set_state(self, state: dict) -> None:
        values = {'expanded_boxes': []}
        values.update(state)
        expanded_boxes = values['expanded_boxes']
        self._set_expanded_boxes(expanded_boxes)

    def set_values(
        self,
        values: dict,
        widgets: dict[str, ParameterWidget] | None = None,
        attr: str = 'value',
    ) -> None:
        if widgets is None:
            widgets = self.widgets()
        for key, value in values.items():
            if key not in widgets:
                continue
            widget = widgets[key]
            if isinstance(widget, dict):
                self.set_values(value, widget, attr)
            else:
                try:
                    setter = getattr(widget, f'set_{attr}')
                except AttributeError:
                    continue
                setter(value)

    def state(self) -> dict:
        expanded_boxes = self._expanded_boxes()
        state = {'expanded_boxes': expanded_boxes}
        return state

    def values(self) -> dict[str, typing.Any]:
        # create nested dictionary of all Parameter values
        values = {}
        for name, widget in self._widgets.items():
            if isinstance(widget, ParameterForm):
                if widget.create_hierarchy:
                    values[name] = widget.values()
                else:
                    values.update(widget.values())
            else:
                values[name] = widget.value()
        return values

    def widgets(self) -> dict[str, ParameterWidget]:
        # create nested dictionary of all Parameter widgets
        widgets = {}
        for name, widget in self._widgets.items():
            if isinstance(widget, ParameterForm):
                if widget.create_hierarchy:
                    widgets[name] = widget.widgets()
                else:
                    widgets.update(widget.widgets())
            else:
                widgets[name] = widget
        return widgets

    def _expanded_boxes(self, groups: dict | None = None) -> list[str]:
        # returns a list of all expanded boxes
        # child boxes are separated by / 'parent/child/grand-child'
        if groups is None:
            groups = self.groups()

        expanded_boxes = []
        for group, children in groups.items():
            if isinstance(group, ParameterBox):
                title = group.title()
                if not group.collapsed():
                    expanded_boxes.append(group.title())
            elif isinstance(group, ParameterForm):
                title = group.name
            else:
                continue

            child_boxes = self._expanded_boxes(children)
            child_boxes = ['/'.join([title, child]) for child in child_boxes]
            expanded_boxes.extend(child_boxes)
        return expanded_boxes

    def _create_form(self, name) -> ParameterForm:
        self._validate_name(name)

        form = ParameterForm(name=name, root=self)
        self._widgets[name] = form

        return form

    def _hierarchical_names(self) -> list[str]:
        # generate a flat list of all child widget names
        names = []
        for name, widget in self._widgets.items():
            if isinstance(widget, ParameterForm):
                names.extend(widget.hierarchical_names())
            else:
                names.append(name)
        return names

    def _set_expanded_boxes(
        self, expanded_boxes: list[str], groups: dict | None = None
    ) -> None:
        # collapses the boxes in the list, child boxes are separated by dot 'parent.child'

        if groups is None:
            groups = self.groups()

        for group, children in groups.items():
            if isinstance(group, ParameterBox):
                group.set_collapsed(group.title() not in expanded_boxes)
                title = group.title()
            elif isinstance(group, ParameterForm):
                title = group.name
            else:
                continue

            child_boxes = [
                b.split('.', 1)[-1]
                for b in expanded_boxes
                if b.split('/', 1)[0] == title
            ]
            self._set_expanded_boxes(child_boxes, children)

    # noinspection PyMethodMayBeStatic
    def _set_widget_row_enabled(self, widget: QtWidgets.QWidget, enabled: bool) -> None:
        # get parent grid layout
        layout = widget.parentWidget().layout()
        if not isinstance(layout, QtWidgets.QGridLayout):
            return

        # find row of widget
        index = layout.indexOf(widget)
        if index < 0:
            return
        row, column, row_span, col_span = layout.getItemPosition(index)

        # widget
        item = layout.itemAtPosition(row, 2)
        if not item or not item.widget():
            return

        item_widget = item.widget()
        item_widget.setEnabled(enabled)

    def _update_stretch(self) -> None:
        layout = self.grid_layout
        layout.setRowStretch(layout.rowCount() - 1, 0)
        layout.setRowStretch(layout.rowCount(), 1)

    def _validate_name(self, name: str) -> None:
        if name is None:
            raise ValueError(f'Cannot add widget with name {name}')

        # check if name is unique relative to root
        if self.root.unique_hierarchical_names:
            hierarchical_names = self.root._hierarchical_names()
        else:
            hierarchical_names = list(self._widgets.keys())
        if name in hierarchical_names:
            raise ValueError(f'Cannot add widget {name} (name already exists)')


class ParameterEditor(ParameterForm):
    def __init__(
        self,
        name: str | None = None,
        root: ParameterEditor | None = None,
        parent: QtWidgets.QWidget | None = None,
    ) -> None:
        super().__init__(name, root, parent)

        self.setLayout(QtWidgets.QVBoxLayout())
        scroll_area = VerticalScrollArea(self)
        self.layout().addWidget(scroll_area)
        self.layout().setContentsMargins(QtCore.QMargins())

        self._form_widget = QtWidgets.QWidget(self)
        self._form_widget.setLayout(QtWidgets.QGridLayout())
        scroll_area.setWidget(self._form_widget)

    @property
    def grid_layout(self) -> QtWidgets.QGridLayout:
        layout = self._form_widget.layout()
        if not isinstance(layout, QtWidgets.QGridLayout):
            raise RuntimeError('Layout needs to be QGridLayout')
        return layout


__all__ = ['ParameterEditor', 'ParameterBox', 'ParameterForm', 'ParameterTabWidget']
